# -*- coding: utf-8 -*-
"""
    emmett.asgi.workers
    -------------------

    Provides ASGI gunicorn workers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio
import logging
import signal

from gunicorn.workers.base import Worker as _Worker

from ..extensions import Signals
from .loops import loops
from .protocols import protocols_http, protocols_ws
from .server import Config, Server


class Worker(_Worker):
    EMMETT_CONFIG = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        logger = logging.getLogger("uvicorn.error")
        logger.handlers = self.log.error_log.handlers
        logger.setLevel(self.log.error_log.level)
        logger.propagate = False

        logger = logging.getLogger("uvicorn.access")
        logger.handlers = self.log.access_log.handlers
        logger.setLevel(self.log.access_log.level)
        logger.propagate = False

        config = {
            "app": None,
            "log_config": None,
            "timeout_keep_alive": self.cfg.keepalive,
            "timeout_notify": self.timeout,
            "callback_notify": self.callback_notify,
            "limit_max_requests": self.max_requests,
            "forwarded_allow_ips": self.cfg.forwarded_allow_ips
        }

        if self.cfg.is_ssl:
            config.update(
                ssl_keyfile=self.cfg.ssl_options.get("keyfile"),
                ssl_certfile=self.cfg.ssl_options.get("certfile"),
                ssl_version=self.cfg.ssl_options.get("ssl_version"),
                ssl_cert_reqs=self.cfg.ssl_options.get("cert_reqs"),
                ssl_ca_certs=self.cfg.ssl_options.get("ca_certs"),
                ssl_ciphers=self.cfg.ssl_options.get("ciphers")
            )

        if self.cfg.settings["backlog"].value:
            config["backlog"] = self.cfg.settings["backlog"].value

        config.update(self.EMMETT_CONFIG)
        config.update(
            http=protocols_http.get(config.get('http', 'auto')),
            ws=protocols_ws.get(config.get('ws', 'auto'))
        )

        self.config = Config(**config)

    def init_process(self):
        self.config.loop = loops.get(self.config.loop)
        super().init_process()

    def init_signals(self):
        for s in self.SIGNALS:
            signal.signal(s, signal.SIG_DFL)

    def run(self):
        self.config.app = self.wsgi
        self.config.app.send_signal(Signals.after_loop, loop=self.config.loop)
        server = Server(config=self.config)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(server.serve(sockets=self.sockets))

    async def callback_notify(self):
        self.notify()


class EmmettWorker(Worker):
    EMMETT_CONFIG = {
        "loop": "uvloop",
        "http": "httptools",
        "proxy_headers": False,
        "interface": "asgi3"
    }


class EmmettH11Worker(EmmettWorker):
    EMMETT_CONFIG = {
        "loop": "asyncio",
        "http": "h11",
        "proxy_headers": False,
        "interface": "asgi3"
    }



# -*- coding: utf-8 -*-
"""
    emmett.asgi.loops
    -----------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..helpers import BuilderRegistry

loops = BuilderRegistry()

from . import (
    asyncio,
    auto,
    uvloop
)



# -*- coding: utf-8 -*-
"""
    emmett.sessions
    ---------------

    Provides session managers for applications.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import os
import pickle
import tempfile
import time
import zlib

from typing import Any, Dict, Optional, Type, TypeVar

from ._internal import warn_of_deprecation
from .ctx import current
from .datastructures import sdict, SessionData
from .pipeline import Pipe
from .security import secure_loads, secure_dumps, uuid
from .wrappers import ScopeWrapper

try:
    from emmett_crypto import symmetric as crypto_symmetric
except Exception:
    crypto_symmetric = None


class SessionPipe(Pipe):
    def __init__(
        self,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None
    ):
        self.expire = expire
        self.secure = secure
        self.samesite = samesite
        self.domain = domain
        self.cookie_name = (
            cookie_name or f'emt_session_data_{current.app.name}'
        )
        self.cookie_data = cookie_data or {}

    def _load_session(self, wrapper: ScopeWrapper):
        raise NotImplementedError

    def _new_session(self) -> SessionData:
        raise NotImplementedError

    def _pack_session(self, expiration: int):
        current.response.cookies[self.cookie_name] = self._session_cookie_data()
        cookie_data = current.response.cookies[self.cookie_name]
        cookie_data['path'] = "/"
        cookie_data['expires'] = expiration
        cookie_data['samesite'] = self.samesite
        if self.secure:
            cookie_data['secure'] = True
        if self.domain is not None:
            cookie_data['domain'] = self.domain
        for key, val in self.cookie_data.items():
            cookie_data[key] = val

    def _session_cookie_data(self) -> str:
        raise NotImplementedError

    async def open_request(self):
        if self.cookie_name in current.request.cookies:
            current.session = self._load_session(current.request)
        if not current.session:
            current.session = self._new_session()

    async def open_ws(self):
        if self.cookie_name in current.websocket.cookies:
            current.session = self._load_session(current.websocket)
        if not current.session:
            current.session = self._new_session()

    async def close_request(self):
        expiration = current.session._expiration or self.expire
        self._pack_session(expiration)

    def clear(self):
        pass


class CookieSessionPipe(SessionPipe):
    def __init__(
        self,
        key,
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None,
        encryption_mode="legacy",
        compression_level=0
    ):
        super().__init__(
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data
        )
        self.key = key
        if encryption_mode == "legacy":
            warn_of_deprecation("legacy encryption_mode", "modern", stack=5)
            self._encrypt_data = self._encrypt_data_legacy
            self._decrypt_data = self._decrypt_data_legacy
        elif encryption_mode == "modern":
            if not crypto_symmetric:
                raise RuntimeError(
                    "You need emmett-crypto to use modern encryption mode"
                )
            self._encrypt_data = self._encrypt_data_modern
            self._decrypt_data = self._decrypt_data_modern
        else:
            raise ValueError("Invalid encryption_mode")
        self.compression_level = compression_level

    def _encrypt_data_legacy(self) -> str:
        return secure_dumps(
            sdict(current.session),
            self.key,
            compression_level=self.compression_level
        )

    def _encrypt_data_modern(self) -> str:
        data = pickle.dumps(sdict(current.session))
        if self.compression_level:
            data = zlib.compress(data, self.compression_level)
        return crypto_symmetric.encrypt_b64(data, self.key)

    def _decrypt_data_legacy(self, data: str) -> SessionData:
        return SessionData(
            secure_loads(data, self.key, compression_level=self.compression_level),
            expires=self.expire
        )

    def _decrypt_data_modern(self, data: str) -> SessionData:
        try:
            ddata = crypto_symmetric.decrypt_b64(data, self.key)
            if self.compression_level:
                ddata = zlib.decompress(ddata)
            rv = pickle.loads(ddata)
        except Exception:
            rv = None
        return SessionData(rv, expires=self.expire)

    def _load_session(self, wrapper: ScopeWrapper) -> SessionData:
        cookie_data = wrapper.cookies[self.cookie_name].value
        return self._decrypt_data(cookie_data)

    def _new_session(self) -> SessionData:
        return SessionData(expires=self.expire)

    def _session_cookie_data(self) -> str:
        return self._encrypt_data()

    def clear(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support sessions clearing. "
            f"Change the 'key' parameter to invalidate existing ones."
        )


class BackendStoredSessionPipe(SessionPipe):
    def _new_session(self):
        return SessionData(sid=uuid())

    def _session_cookie_data(self) -> str:
        return current.session._sid

    def _load_session(self, wrapper: ScopeWrapper) -> Optional[SessionData]:
        sid = wrapper.cookies[self.cookie_name].value
        data = self._load(sid)
        if data is not None:
            return SessionData(data, sid=sid)
        return None

    def _delete_session(self):
        pass

    def _save_session(self, expiration: int):
        pass

    def _load(self, sid: str):
        return None

    async def close_request(self):
        if not current.session:
            self._delete_session()
            if current.session._modified:
                #: if we got here means we want to destroy session definitely
                if self.cookie_name in current.response.cookies:
                    del current.response.cookies[self.cookie_name]
            return
        expiration = current.session._expiration or self.expire
        self._save_session(expiration)
        self._pack_session(expiration)


class FileSessionPipe(BackendStoredSessionPipe):
    _fs_transaction_suffix = '.__emt_sess'
    _fs_mode = 0o600

    def __init__(
        self,
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None,
        filename_template='emt_%s.sess'
    ):
        super().__init__(
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data
        )
        assert not filename_template.endswith(self._fs_transaction_suffix), \
            'filename templates cannot end with %s' % \
            self._fs_transaction_suffix
        self._filename_template = filename_template
        self._path = os.path.join(current.app.root_path, 'sessions')
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)

    def _delete_session(self):
        fn = self._get_filename(current.session._sid)
        try:
            os.unlink(fn)
        except OSError:
            pass

    def _save_session(self, expiration):
        if current.session._modified:
            self._store(current.session, expiration)

    def _get_filename(self, sid):
        return os.path.join(self._path, self._filename_template % str(sid))

    def _load(self, sid):
        try:
            with open(self._get_filename(sid), 'rb') as f:
                exp = pickle.load(f)
                val = pickle.load(f)
        except IOError:
            return None
        if exp < time.time():
            return None
        return val

    def _store(self, session, expiration):
        fn = self._get_filename(session._sid)
        now = time.time()
        exp = now + expiration
        fd, tmp = tempfile.mkstemp(
            suffix=self._fs_transaction_suffix, dir=self._path)
        f = os.fdopen(fd, 'wb')
        try:
            pickle.dump(exp, f, 1)
            pickle.dump(sdict(session), f, pickle.HIGHEST_PROTOCOL)
        finally:
            f.close()
        try:
            os.rename(tmp, fn)
            os.chmod(fn, self._fs_mode)
        except Exception:
            pass

    def clear(self):
        for element in os.listdir(self._path):
            try:
                os.unlink(os.path.join(self._path, element))
            except Exception:
                pass


class RedisSessionPipe(BackendStoredSessionPipe):
    def __init__(
        self,
        redis,
        prefix="emtsess:",
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None
    ):
        super().__init__(
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data
        )
        self.redis = redis
        self.prefix = prefix

    def _delete_session(self):
        self.redis.delete(self.prefix + current.session._sid)

    def _save_session(self, expiration):
        if current.session._modified:
            self.redis.setex(
                self.prefix + current.session._sid,
                expiration,
                current.session._dump
            )
        else:
            self.redis.expire(self.prefix + current.session._sid, expiration)

    def _load(self, sid):
        data = self.redis.get(self.prefix + sid)
        return pickle.loads(data) if data else data

    def clear(self):
        self.redis.delete(self.prefix + "*")


TSessionPipe = TypeVar("TSessionPipe", bound=SessionPipe)


class SessionManager:
    _pipe: Optional[SessionPipe] = None

    @classmethod
    def _build_pipe(
        cls,
        handler_cls: Type[TSessionPipe],
        *args: Any,
        **kwargs: Any
    ) -> TSessionPipe:
        cls._pipe = handler_cls(*args, **kwargs)
        return cls._pipe

    @classmethod
    def cookies(
        cls,
        key: str,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
        encryption_mode: str = "legacy",
        compression_level: int = 0
    ) -> CookieSessionPipe:
        return cls._build_pipe(
            CookieSessionPipe,
            key,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
            encryption_mode=encryption_mode,
            compression_level=compression_level
        )

    @classmethod
    def files(
        cls,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
        filename_template: str = 'emt_%s.sess'
    ) -> FileSessionPipe:
        return cls._build_pipe(
            FileSessionPipe,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
            filename_template=filename_template
        )

    @classmethod
    def redis(
        cls,
        redis: Any,
        prefix: str = "emtsess:",
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None
    ) -> RedisSessionPipe:
        return cls._build_pipe(
            RedisSessionPipe,
            redis,
            prefix=prefix,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data
        )

    @classmethod
    def clear(cls):
        cls._pipe.clear()



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.base
    --------------------------

    Provides base migrations objects.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ...datastructures import sdict
from .. import Model, Field
from .engine import MetaEngine, Engine
from .helpers import WrappedOperation, _feasible_as_dbms_default


class Schema(Model):
    tablename = "emmett_schema"
    version = Field()


class Migration(object):
    _registered_ops_ = {}

    @classmethod
    def register_operation(cls, name):
        def wrap(op_cls):
            cls._registered_ops_[name] = op_cls
            return op_cls
        return wrap

    def __init__(self, app, db, is_meta=False):
        self.db = db
        if is_meta:
            self.engine = MetaEngine(db)
        else:
            self.engine = Engine(db)

    def __getattr__(self, name):
        registered = self._registered_ops_.get(name)
        if registered is not None:
            return WrappedOperation(registered, name, self.engine)
        else:
            raise NotImplementedError


class Column(sdict):
    def __init__(self, name, type='string', unique=False, notnull=False,
                 **kwargs):
        self.name = name
        self.type = type
        self.unique = unique
        self.notnull = notnull
        for key, val in kwargs.items():
            self[key] = val
        self.length = self.length or 255

    def _build_fks(self, db, tablename):
        if self.type.startswith('reference'):
            referenced = self.type[10:].strip()
            try:
                rtablename, rfieldname = referenced.split('.')
            except Exception:
                rtablename = referenced
                rfieldname = 'id'
            if not rtablename:
                rtablename = tablename
            rtable = db[rtablename]
            rfield = rtable[rfieldname]
            if getattr(rtable, '_primarykey', None) and rfieldname in \
                    rtable._primarykey or rfield.unique:
                if not rfield.unique and len(rtable._primarykey) > 1:
                    # self.tfk = [pk for pk in rtable._primarykey]
                    raise NotImplementedError(
                        'Column of type reference pointing to multiple ' +
                        'columns are currently not supported.'
                    )
                else:
                    self.fk = True

    @classmethod
    def from_field(cls, field):
        rv = cls(
            field.name,
            field.type,
            field.unique,
            field.notnull,
            length=field.length,
            ondelete=field.ondelete
        )
        if _feasible_as_dbms_default(field.default):
            rv.default = field.default
        rv._build_fks(field.db, field.tablename)
        return rv

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(["%s=%r" % (k, v) for k, v in self.items()])
        )



# -*- coding: utf-8 -*-
"""
    emmett.routing.router
    ---------------------

    Provides router implementations.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import re

from collections import namedtuple
from typing import Any, Callable, Dict, List, Type

from ..ctx import current
from ..extensions import Signals
from ..http import HTTP
from .response import (
    MetaResponseBuilder,
    EmptyResponseBuilder,
    ResponseBuilder,
    AutoResponseBuilder,
    BytesResponseBuilder,
    TemplateResponseBuilder
)
from .rules import RoutingRule, HTTPRoutingRule, WebsocketRoutingRule


RouteRecReq = namedtuple(
    "RouteRecReq", ["name", "match", "dispatch"]
)
RouteRecWS = namedtuple(
    "RouteRecWS", ["name", "match", "dispatch", "flow_recv", "flow_send"]
)


class Router:
    __slots__ = [
        '_get_routes_in_for_host',
        '_match_lang',
        '_prefix_main_len',
        '_prefix_main',
        '_routes_nohost',
        '_routes_str',
        'app',
        'routes_in',
        'routes_out',
        'routes'
    ]

    _outputs: Dict[str, Type[MetaResponseBuilder]] = {}
    _routing_rule_cls: Type[RoutingRule] = RoutingRule
    _routing_signal = Signals.before_routes
    _routing_started = False
    _routing_stack: List[RoutingRule] = []
    _re_components = re.compile(r'(\()?([^<\w]+)?<(\w+)\:(\w+)>(\)\?)?')

    def __init__(self, app, url_prefix=None):
        self.app = app
        self.routes = []
        self.routes_in = {'__any__': self._build_routing_dict()}
        self.routes_out = {}
        self._routes_str = {}
        self._routes_nohost = (self.routes_in['__any__'], )
        self._get_routes_in_for_host = self._get_routes_in_for_host_nomatch
        main_prefix = url_prefix or ''
        if main_prefix:
            main_prefix = main_prefix.rstrip('/')
            if not main_prefix.startswith('/'):
                main_prefix = '/' + main_prefix
            if main_prefix == '/':
                main_prefix = ''
        self._prefix_main = main_prefix
        self._prefix_main_len = len(self._prefix_main)
        self._set_language_handling()

    def _set_language_handling(self):
        self._match_lang = (
            self._match_with_lang if self.app.language_force_on_url
            else self._match_no_lang)

    @property
    def static_versioning(self):
        return (
            self.app.config.static_version_urls and
            self.app.config.static_version
        ) or ''

    @staticmethod
    def _build_routing_dict():
        return {'static': {}, 'match': {}}

    @classmethod
    def build_route_components(cls, path):
        components = []
        params = []
        for match in cls._re_components.findall(path):
            params.append(match[1] + "{}")
        statics = cls._re_components.sub("{}", path).split("{}")
        if not params:
            components = statics
        else:
            components.append(statics[0])
            for idx, _ in enumerate(params):
                components.append(params[idx] + statics[idx + 1])
        return components

    def _get_routes_in_for_host_match(self, wrapper):
        return (
            self.routes_in.get(wrapper.host, self.routes_in['__any__']),
            self.routes_in['__any__'])

    def _get_routes_in_for_host_nomatch(self, wrapper):
        return self._routes_nohost

    def _match_with_lang(self, wrapper, path):
        path, lang = self._split_lang(path)
        current.language = wrapper.language = lang
        return path

    def _match_no_lang(self, wrapper, path):
        wrapper.language = None
        return path

    @staticmethod
    def remove_trailslash(path):
        return path.rstrip('/') or path

    def _split_lang(self, path):
        default = self.app.language_default
        if len(path) <= 1:
            return path, default
        clean_path = path.lstrip('/')
        if clean_path[2:3] == '/':
            lang, new_path = clean_path[:2], clean_path[2:]
            if lang != default and lang in self.app._languages_set:
                return new_path, lang
        return path, default

    def add_route(self, route):
        raise NotImplementedError

    def match(self, wrapper):
        raise NotImplementedError

    async def dispatch(self):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        if not self.__class__._routing_started:
            self.__class__._routing_started = True
            self.app.send_signal(self._routing_signal)
        return RoutingCtx(self, self._routing_rule_cls, *args, **kwargs)

    @classmethod
    def exposing(cls):
        return cls._routing_stack[-1]


class HTTPRouter(Router):
    __slots__ = ['pipeline', 'injectors']

    _routing_rule_cls = HTTPRoutingRule
    _routing_rec_builder = RouteRecReq

    _outputs = {
        'empty': EmptyResponseBuilder,
        'auto': AutoResponseBuilder,
        'bytes': BytesResponseBuilder,
        'str': ResponseBuilder,
        'template': TemplateResponseBuilder
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = []
        self.injectors = []

    @staticmethod
    def _build_routing_dict():
        rv = {}
        for scheme in ['http', 'https']:
            rv[scheme] = {}
            for method in [
                'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT'
            ]:
                rv[scheme][method] = {'static': {}, 'match': {}}
        return rv

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s %s://%s%s%s -> %s" % (
            "|".join(route.methods),
            "|".join(route.schemes),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name
        )

    def add_route(self, route):
        self.routes.append(route)
        host = route.hostname or '__any__'
        if host not in self.routes_in:
            self.routes_in[host] = self._build_routing_dict()
            self._get_routes_in_for_host = self._get_routes_in_for_host_match
        for scheme in route.schemes:
            for method in route.methods:
                routing_dict = self.routes_in[host][scheme][method]
                slot, key = (
                    ('static', route.path) if route.is_static else
                    ('match', route.name)
                )
                routing_dict[slot][key] = self._routing_rec_builder(
                    name=route.name,
                    match=route.match,
                    dispatch=route.dispatchers[method].dispatch
                )
        self.routes_out[route.name] = {
            'host': route.hostname,
            'path': self.build_route_components(route.path)
        }
        self.add_route_str(route)

    def match(self, request):
        path = self._match_lang(
            request,
            self.remove_trailslash(request.path)
        )
        for routing_dict in self._get_routes_in_for_host(request):
            sub_dict = routing_dict[request.scheme][request.method]
            element = sub_dict['static'].get(path)
            if element:
                return element, {}
            for element in sub_dict['match'].values():
                match, args = element.match(path)
                if match:
                    return element, args
        return None, {}

    async def dispatch(self, request, response):
        match, reqargs = self.match(request)
        if not match:
            raise HTTP(404, body="Resource not found\n")
        request.name = match.name
        return await match.dispatch(reqargs, response)


class WebsocketRouter(Router):
    __slots__ = ['pipeline']

    _routing_rule_cls = WebsocketRoutingRule
    _routing_rec_builder = RouteRecWS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = []

    @staticmethod
    def _build_routing_dict():
        rv = {}
        for scheme in ['ws', 'wss']:
            rv[scheme] = {'static': {}, 'match': {}}
        return rv

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s://%s%s%s -> %s" % (
            "|".join(route.schemes),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name
        )

    def add_route(self, route):
        self.routes.append(route)
        host = route.hostname or '__any__'
        if host not in self.routes_in:
            self.routes_in[host] = self._build_routing_dict()
            self._get_routes_in_for_host = self._get_routes_in_for_host_match
        for scheme in route.schemes:
            routing_dict = self.routes_in[host][scheme]
            slot, key = (
                ('static', route.path) if route.is_static else
                ('match', route.name)
            )
            routing_dict[slot][key] = self._routing_rec_builder(
                name=route.name,
                match=route.match,
                dispatch=route.dispatcher.dispatch,
                flow_recv=route.pipeline_flow_receive,
                flow_send=route.pipeline_flow_send
            )
        self.routes_out[route.name] = {
            'host': route.hostname,
            'path': self.build_route_components(route.path)
        }
        self.add_route_str(route)

    def match(self, websocket):
        path = self._match_lang(
            websocket,
            self.remove_trailslash(websocket.path)
        )
        for routing_dict in self._get_routes_in_for_host(websocket):
            sub_dict = routing_dict[websocket.scheme]
            element = sub_dict['static'].get(path)
            if element:
                return element, {}
            for element in sub_dict['match'].values():
                match, args = element.match(path)
                if match:
                    return element, args
        return None, {}

    async def dispatch(self, websocket):
        match, reqargs = self.match(websocket)
        if not match:
            raise HTTP(404, body="Resource not found\n")
        websocket.name = match.name
        websocket._bind_flow(
            match.flow_recv,
            match.flow_send
        )
        await match.dispatch(reqargs)


class RoutingCtx:
    __slots__ = ['router', 'rule']

    def __init__(
        self,
        router: Router,
        rule_cls: Type[RoutingRule],
        *args,
        **kwargs
    ):
        self.router = router
        self.rule = rule_cls(self.router, *args, **kwargs)
        self.router._routing_stack.append(self.rule)

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        self.router.app.send_signal(Signals.before_route, route=self.rule, f=f)
        rv = self.rule(f)
        self.router.app.send_signal(Signals.after_route, route=self.rule)
        self.router._routing_stack.pop()
        return rv



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.request
    -----------------------

    Provides http request wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio
import re

from cgi import FieldStorage, parse_header
from datetime import datetime
from io import BytesIO
from urllib.parse import parse_qs
from typing import Any, Optional

import pendulum

from ..asgi.typing import Scope, Receive, Send
from ..datastructures import sdict
from ..http import HTTP
from ..parsers import Parsers
from ..utils import cachedprop
from . import ScopeWrapper
from .helpers import FileStorage, RequestCancelled

_regex_client = re.compile(r'[\w\-:]+(\.[\w\-]+)*\.?')
_push_headers = {
    "accept",
    "accept-encoding",
    "accept-language",
    "cache-control",
    "user-agent"
}


class Body:
    __slots__ = ('_data', '_receive', '_max_content_length')

    def __init__(self, receive, max_content_length=None):
        self._data = bytearray()
        self._receive = receive
        self._max_content_length = max_content_length

    def append(self, data: bytes):
        if data == b'':
            return
        self._data.extend(data)
        if (
            self._max_content_length is not None and
            len(self._data) > self._max_content_length
        ):
            raise HTTP(413, 'Request entity too large')

    async def __load(self) -> bytes:
        while True:
            event = await self._receive()
            if event['type'] == 'http.request':
                self.append(event['body'])
                if not event.get('more_body', False):
                    break
            elif event['type'] == 'http.disconnect':
                raise RequestCancelled
        return bytes(self._data)

    def __await__(self):
        return self.__load().__await__()


class Request(ScopeWrapper):
    __slots__ = ['_now', 'method']

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        max_content_length: Optional[int] = None,
        body_timeout: Optional[int] = None
    ):
        super().__init__(scope, receive, send)
        self.max_content_length = max_content_length
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope['method']

    @cachedprop
    def _input(self):
        return Body(self._receive, self.max_content_length)

    @cachedprop
    async def body(self) -> bytes:
        if (
            self.max_content_length and
            self.content_length > self.max_content_length
        ):
            raise HTTP(413, 'Request entity too large')
        try:
            rv = await asyncio.wait_for(self._input, timeout=self.body_timeout)
        except asyncio.TimeoutError:
            raise HTTP(408, 'Request timeout')
        return rv

    @cachedprop
    def now(self) -> pendulum.DateTime:
        return pendulum.instance(self._now)

    @cachedprop
    def now_local(self) -> pendulum.DateTime:
        return self.now.in_timezone(pendulum.local_timezone())  # type: ignore

    @cachedprop
    def content_type(self) -> str:
        return parse_header(self.headers.get('content-type', ''))[0]

    @cachedprop
    def content_length(self) -> int:
        return self.headers.get('content_length', 0, cast=int)

    _empty_body_methods = {v: v for v in ['GET', 'HEAD', 'OPTIONS']}

    @cachedprop
    async def _input_params(self):
        if self._empty_body_methods.get(self.method) or not self.content_type:
            return sdict(), sdict()
        return await self._load_params()

    @cachedprop
    async def body_params(self) -> sdict[str, Any]:
        rv, _ = await self._input_params
        return rv

    @cachedprop
    async def files(self) -> sdict[str, FileStorage]:
        _, rv = await self._input_params
        return rv

    def _load_params_missing(self, data):
        return sdict(), sdict()

    def _load_params_json(self, data):
        try:
            params = Parsers.get_for('json')(data)
        except Exception:
            params = {}
        return sdict(params), sdict()

    def _load_params_form_urlencoded(self, data):
        rv = sdict()
        for key, values in parse_qs(
            data.decode('latin-1'), keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv, sdict()

    def _load_params_form_multipart(self, data):
        params, files = sdict(), sdict()
        field_storage = FieldStorage(
            BytesIO(data),
            headers=self.headers,
            environ={'REQUEST_METHOD': self.method},
            keep_blank_values=True
        )
        for key in field_storage:
            field = field_storage[key]
            if isinstance(field, list):
                if len(field) > 1:
                    params[key] = []
                    for element in field:
                        params[key].append(element.value)
                else:
                    params[key] = field[0].value
            elif (
                isinstance(field, FieldStorage) and
                field.filename is not None
            ):
                files[key] = FileStorage(
                    BytesIO(field.file.read()),
                    field.filename,
                    field.name,
                    field.type,
                    field.headers
                )
                continue
            else:
                params[key] = field.value
        return params, files

    _params_loaders = {
        'application/json': _load_params_json,
        'application/x-www-form-urlencoded': _load_params_form_urlencoded,
        'multipart/form-data': _load_params_form_multipart
    }

    async def _load_params(self):
        loader = self._params_loaders.get(
            self.content_type, self._load_params_missing)
        return loader(self, await self.body)

    @cachedprop
    def client(self) -> str:
        g = _regex_client.search(self.headers.get('x-forwarded-for', ''))
        client = (g.group() or '').split(',')[0] if g else None
        if client in (None, '', 'unknown'):
            g = _regex_client.search(self.headers.get('remote-addr', ''))
            if g:
                client = g.group()
            elif self.host.startswith('['):
                # IPv6
                client = '::1'
            else:
                # IPv4
                client = '127.0.0.1'
        return client  # type: ignore

    async def push_promise(self, path: str):
        if "http.response.push" not in self._scope.get("extensions", {}):
            return
        await self._send({
            "type": "http.response.push",
            "path": path,
            "headers": [
                (key.encode("latin-1"), self.headers[key].encode("latin-1"))
                for key in _push_headers & set(self.headers.keys())
            ]
        })



# -*- coding: utf-8 -*-
"""
    emmett.orm.objects
    ------------------

    Provides pyDAL objects implementation for Emmett.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import copy
import datetime
import decimal
import types

from collections import OrderedDict, defaultdict
from pydal.objects import (
    Table as _Table, Field as _Field, Set as _Set,
    Row as _Row, Rows as _Rows, IterRows as _IterRows, Query, Expression)

from ..ctx import current
from ..datastructures import sdict
from ..html import tag
from ..serializers import xml_encode
from ..utils import cachedprop
from ..validators import ValidateFromDict
from .helpers import (
    _IDReference, JoinedIDReference, RelationBuilder, wrap_scope_on_set)


class Table(_Table):
    def __init__(self, *args, **kwargs):
        super(Table, self).__init__(*args, **kwargs)
        self._unique_fields_validation_ = {}

    def _create_references(self):
        self._referenced_by = []
        self._referenced_by_list = []
        self._references = []


class Field(_Field):
    _internal_types = {
        'integer': 'int',
        'double': 'float',
        'boolean': 'bool',
        'list:integer': 'list:int'
    }
    _pydal_types = {
        'int': 'integer',
        'bool': 'boolean',
        'list:int': 'list:integer',
    }
    _internal_delete = {
        'cascade': 'CASCADE', 'nullify': 'SET NULL', 'nothing': 'NO ACTION'
    }
    _inst_count_ = 0
    _obj_created_ = False

    def __init__(self, type='string', *args, **kwargs):
        self.modelname = None
        #: convert type
        self._type = self._internal_types.get(type, type)
        #: convert 'rw' -> 'readable', 'writeable'
        if 'rw' in kwargs:
            _rw = kwargs.pop('rw')
            if isinstance(_rw, (tuple, list)):
                read, write = _rw
            else:
                read = write = _rw
            kwargs['readable'] = read
            kwargs['writable'] = write
        #: convert 'info' -> 'comment'
        _info = kwargs.pop('info', None)
        if _info:
            kwargs['comment'] = _info
        #: convert ondelete parameter
        _ondelete = kwargs.get('ondelete')
        if _ondelete:
            if _ondelete not in list(self._internal_delete):
                raise SyntaxError(
                    'Field ondelete should be set on %s, %s or %s' %
                    list(self._internal_delete)
                )
            kwargs['ondelete'] = self._internal_delete[_ondelete]
        #: process 'refers_to' fields
        self._isrefers = kwargs.pop('_isrefers', None)
        #: get auto validation preferences
        self._auto_validation = kwargs.pop('auto_validation', True)
        #: intercept validation (will be processed by `_make_field`)
        self._requires = {}
        self._custom_requires = []
        if 'validation' in kwargs:
            _validation = kwargs.pop('validation')
            if isinstance(_validation, dict):
                self._requires = _validation
            else:
                self._custom_requires = _validation
                if not isinstance(self._custom_requires, list):
                    self._custom_requires = [self._custom_requires]
        self._validation = {}
        self._vparser = ValidateFromDict()
        #: ensure 'length' is an integer
        if 'length' in kwargs:
            kwargs['length'] = int(kwargs['length'])
        #: store args and kwargs for `_make_field`
        self._args = args
        self._kwargs = kwargs
        #: increase creation counter (used to keep order of fields)
        self._inst_count_ = Field._inst_count_
        Field._inst_count_ += 1

    def _default_validation(self):
        rv = {}
        auto_types = [
            'int', 'float', 'date', 'time', 'datetime', 'json'
        ]
        if self._type in auto_types:
            rv['is'] = self._type
        elif self._type.startswith('decimal'):
            rv['is'] = 'decimal'
        if self._type == 'bigint':
            rv['is'] = 'int'
        if self._type == 'bool':
            rv['in'] = (False, True)
        if self._type in ['string', 'text', 'password']:
            rv['len'] = {'lte': self.length}
        if self._type == 'password':
            rv['len']['gte'] = 6
            rv['crypt'] = True
        if self._type == 'list:int':
            rv['is'] = 'list:int'
        if (
            self.notnull or self._type.startswith('reference') or
            self._type.startswith('list:reference')
        ):
            rv['presence'] = True
        if not self.notnull and self._isrefers is True:
            rv['allow'] = 'empty'
        if self.unique:
            rv['unique'] = True
        return rv

    def _parse_validation(self):
        for key in list(self._requires):
            self._validation[key] = self._requires[key]
        self.requires = self._vparser(self, self._validation) + \
            self._custom_requires

    #: `_make_field` will be called by `Model` class or `Form` class
    #  it will make intenral Field class compatible with the pyDAL's one
    def _make_field(self, name, model=None):
        if self._obj_created_:
            return self
        if model is not None:
            self.modelname = model.__class__.__name__
        #: convert field type to pyDAL ones if needed
        ftype = self._pydal_types.get(self._type, self._type)
        #: create pyDAL's Field instance
        super(Field, self).__init__(name, ftype, *self._args, **self._kwargs)
        #: add automatic validation (if requested)
        if self._auto_validation:
            auto = True
            if self.modelname:
                auto = model.auto_validation
            if auto:
                self._validation = self._default_validation()
        #: validators
        if not self.modelname:
            self._parse_validation()
        self._obj_created_ = True
        return self

    def __str__(self):
        if self._obj_created_:
            return super(Field, self).__str__()
        return object.__str__(self)

    def __repr__(self):
        if self.modelname and hasattr(self, 'name'):
            return "<%s.%s (%s) field>" % (self.modelname, self.name,
                                           self._type)
        return super(Field, self).__repr__()

    @classmethod
    def string(cls, *args, **kwargs):
        return cls('string', *args, **kwargs)

    @classmethod
    def int(cls, *args, **kwargs):
        return cls('int', *args, **kwargs)

    @classmethod
    def bigint(cls, *args, **kwargs):
        return cls('bigint', *args, **kwargs)

    @classmethod
    def float(cls, *args, **kwargs):
        return cls('float', *args, **kwargs)

    @classmethod
    def text(cls, *args, **kwargs):
        return cls('text', *args, **kwargs)

    @classmethod
    def bool(cls, *args, **kwargs):
        return cls('bool', *args, **kwargs)

    @classmethod
    def blob(cls, *args, **kwargs):
        return cls('blob', *args, **kwargs)

    @classmethod
    def date(cls, *args, **kwargs):
        return cls('date', *args, **kwargs)

    @classmethod
    def time(cls, *args, **kwargs):
        return cls('time', *args, **kwargs)

    @classmethod
    def datetime(cls, *args, **kwargs):
        return cls('datetime', *args, **kwargs)

    @classmethod
    def decimal(cls, precision, scale, *args, **kwargs):
        return cls('decimal({},{})'.format(precision, scale), *args, **kwargs)

    @classmethod
    def json(cls, *args, **kwargs):
        return cls('json', *args, **kwargs)

    @classmethod
    def jsonb(cls, *args, **kwargs):
        return cls('jsonb', *args, **kwargs)

    @classmethod
    def password(cls, *args, **kwargs):
        return cls('password', *args, **kwargs)

    @classmethod
    def upload(cls, *args, **kwargs):
        return cls('upload', *args, **kwargs)

    @classmethod
    def int_list(cls, *args, **kwargs):
        return cls('list:int', *args, **kwargs)

    @classmethod
    def string_list(cls, *args, **kwargs):
        return cls('list:string', *args, **kwargs)

    def cast(self, value, **kwargs):
        return Expression(
            self.db, self._dialect.cast, self,
            self._dialect.types[value] % kwargs, value)


class Set(_Set):
    def __init__(self, db, query, ignore_common_filters=None, model=None):
        super(Set, self).__init__(db, query, ignore_common_filters)
        self._model_ = model
        self._scopes_ = {}
        self._load_scopes_()

    def _load_scopes_(self):
        if self._model_ is None:
            tablemap = self.db._adapter.tables(self.query)
            if len(tablemap) == 1:
                self._model_ = tablemap.popitem()[1]._model_
        if self._model_:
            self._scopes_ = self._model_._instance_()._scopes_

    def _clone(self, ignore_common_filters=None, model=None, **changes):
        return self.__class__(
            self.db, changes.get('query', self.query),
            ignore_common_filters=ignore_common_filters,
            model=model or self._model_
        )

    def where(self, query, ignore_common_filters=None, model=None):
        if query is None:
            return self
        elif isinstance(query, Table):
            query = self.db._adapter.id_query(query)
        elif isinstance(query, str):
            query = Expression(self.db, query)
        elif isinstance(query, Field):
            query = query != None
        elif isinstance(query, types.LambdaType):
            model = model or self._model_
            if not model:
                raise ValueError(
                    "Too many models involved in the Set to use a lambda")
            query = query(model)
        q = self.query & query if self.query else query
        return self._clone(ignore_common_filters, model, query=q)

    def _parse_paginate(self, pagination):
        if isinstance(pagination, tuple):
            offset = pagination[0]
            limit = pagination[1]
        else:
            offset = pagination
            limit = 10
        return ((offset - 1) * limit, offset * limit)

    def _join_set_builder(self, obj, jdata, auto_select_tables):
        return JoinedSet._from_set(
            obj, jdata=jdata, auto_select_tables=auto_select_tables)

    def _left_join_set_builder(self, jdata):
        return JoinedSet._from_set(
            self, ljdata=jdata, auto_select_tables=[self._model_.table])

    def _run_select_(self, *fields, **options):
        return super(Set, self).select(*fields, **options)

    def _get_table_from_query(self):
        if self._model_:
            return self._model_.table
        return self.db._adapter.get_table(self.query)

    def select(self, *fields, **options):
        obj = self
        pagination, including = (
            options.pop('paginate', None), options.pop('including', None))
        if pagination:
            options['limitby'] = self._parse_paginate(pagination)
        if including and self._model_ is not None:
            options['left'], jdata = self._parse_left_rjoins(including)
            obj = self._left_join_set_builder(jdata)
        return obj._run_select_(*fields, **options)

    def update(self, **update_fields):
        table = self._get_table_from_query()
        row = table._fields_and_values_for_update(update_fields)
        if not row._values:
            raise ValueError("No fields to update")
        if any(f(self, row) for f in table._before_update):
            return 0
        ret = self.db._adapter.update(table, self.query, row.op_values())
        ret and [f(self, row) for f in table._after_update]
        return ret

    def validate_and_update(self, **update_fields):
        table = self._get_table_from_query()
        current._dbvalidation_record_id_ = None
        if table._unique_fields_validation_ and self.count() == 1:
            if any(
                table._unique_fields_validation_.get(fieldname)
                for fieldname in update_fields.keys()
            ):
                current._dbvalidation_record_id_ = \
                    self.select(table.id).first().id
        response = Row()
        response.errors = Row()
        new_fields = copy.copy(update_fields)
        for key, value in update_fields.items():
            value, error = table[key].validate(value)
            if error:
                response.errors[key] = '%s' % error
            else:
                new_fields[key] = value
        del current._dbvalidation_record_id_
        if response.errors:
            response.updated = None
        else:
            row = table._fields_and_values_for_update(new_fields)
            if not row._values:
                raise ValueError("No fields to update")
            if any(f(self, row) for f in table._before_update):
                ret = 0
            else:
                ret = self.db._adapter.update(
                    table, self.query, row.op_values())
                ret and [f(self, row) for f in table._after_update]
            response.updated = ret
        return response

    def join(self, *args):
        rv = self
        if self._model_ is not None:
            joins = []
            jdata = []
            auto_select_tables = [self._model_.table]
            for arg in args:
                condition, table, rel_type = self._parse_rjoin(arg)
                joins.append(condition)
                jdata.append((arg, table._tablename, rel_type))
                auto_select_tables.append(table)
            if joins:
                q = joins[0]
                for join in joins[1:]:
                    q = q & join
                rv = rv.where(q, model=self._model_)
                return self._join_set_builder(rv, jdata, auto_select_tables)
        return rv

    def switch(self, model):
        self._model_ = model
        return self

    def _parse_rjoin(self, arg):
        #: match has_many
        rel = self._model_._hasmany_ref_.get(arg)
        if rel:
            if isinstance(rel, dict) and rel.get('via'):
                r = RelationBuilder(rel, self._model_._instance_()).via()
                return r[0], r[1]._table, 'many'
            else:
                r = RelationBuilder(rel, self._model_._instance_())
                return r.many(), rel.table, 'many'
        #: match belongs_to and refers_to
        rel = self._model_._belongs_ref_.get(arg)
        if rel:
            r = RelationBuilder(
                (rel, arg), self._model_._instance_()
            ).belongs_query()
            return r, self._model_.db[rel], 'belongs'
        #: match has_one
        rel = self._model_._hasone_ref_.get(arg)
        if rel:
            r = RelationBuilder(rel, self._model_._instance_())
            return r.many(), rel.table, 'one'
        raise RuntimeError(
            'Unable to find %s relation of %s model' %
            (arg, self._model_.__name__))

    def _parse_left_rjoins(self, args):
        if not isinstance(args, (list, tuple)):
            args = [args]
        joins = []
        jdata = []
        for arg in args:
            condition, table, rel_type = self._parse_rjoin(arg)
            joins.append(table.on(condition))
            jdata.append((arg, table._tablename, rel_type))
        return joins, jdata

    def _jcolnames_from_rowstmps(self, tmps):
        colnames = []
        all_colnames = {}
        jcolnames = {}
        for colname in tmps:
            all_colnames[colname[0]] = colname[0]
            jcolnames[colname[0]] = jcolnames.get(colname[0], [])
            jcolnames[colname[0]].append(colname[1])
        for colname in all_colnames.keys():
            if colname == self._stable_:
                colnames.append(colname)
                del jcolnames[colname]
        return colnames, jcolnames

    def __getattr__(self, name):
        scope = self._scopes_.get(name)
        if scope:
            return wrap_scope_on_set(self, self._model_._instance_(), scope.f)
        raise AttributeError(name)


class RelationSet(object):
    _relation_method_ = 'many'

    def __init__(self, db, relation_builder, row):
        self.db = db
        self._relation_ = relation_builder
        self._row_ = row

    def _get_query_(self):
        try:
            return getattr(self._relation_, self._relation_method_)(self._row_)
        except AttributeError as e:
            raise RuntimeError(e.message)
        except Exception:
            raise

    @property
    def _model_(self):
        return self._relation_.ref.model_instance

    @cachedprop
    def _field_(self):
        return self._relation_.ref.field_instance

    @cachedprop
    def _scopes_(self):
        if self._relation_.ref.method:
            return [lambda: self._relation_.ref.dbset.query]
        return self._relation_._extra_scopes(self._relation_.ref)

    @cachedprop
    def _set(self):
        return Set(self.db, self._get_query_(), model=self._model_.__class__)

    def __getattr__(self, name):
        return getattr(self._set, name)

    def __repr__(self):
        return repr(self._set)

    def __call__(self, query, ignore_common_filters=False):
        return self._set.where(query, ignore_common_filters)

    def _last_resultset(self, refresh=False):
        if refresh or not hasattr(self, '_cached_resultset'):
            self._cached_resultset = self._cache_resultset()
        return self._cached_resultset

    def _filter_reload(self, kwargs):
        return kwargs.pop('reload', False)

    def create(self, **kwargs):
        attributes = self._get_fields_from_scopes(
            self._scopes_, self._model_.tablename)
        attributes.update(**kwargs)
        attributes[self._field_.name] = self._row_.id
        return self._model_.create(
            **attributes
        )

    @staticmethod
    def _get_fields_from_scopes(scopes, table_name):
        rv = {}
        for scope in scopes:
            query = scope()
            components = [query.second, query.first]
            current_kv = []
            while components:
                component = components.pop()
                if isinstance(component, Query):
                    components.append(component.second)
                    components.append(component.first)
                else:
                    if isinstance(component, Field) and \
                       component._tablename == table_name:
                        current_kv.append(component)
                    else:
                        if current_kv:
                            current_kv.append(component)
                        else:
                            components.pop()
                if len(current_kv) > 1:
                    rv[current_kv[0].name] = current_kv[1]
                    current_kv = []
        return rv


class HasOneSet(RelationSet):
    def _cache_resultset(self):
        return self.select(self._model_.table.ALL, limitby=(0, 1)).first()

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args and not kwargs:
            return self._last_resultset(refresh)
        kwargs['limitby'] = (0, 1)
        return self.select(*args, **kwargs).first()


class HasManySet(RelationSet):
    def _cache_resultset(self):
        return self.select(self._model_.table.ALL)

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args and not kwargs:
            return self._last_resultset(refresh)
        return self.select(*args, **kwargs)

    def add(self, obj):
        attributes = self._get_fields_from_scopes(
            self._scopes_, self._model_.tablename)
        attributes[self._field_.name] = self._row_.id
        return self.db(
            self.db[self._field_._tablename].id == obj.id
        ).validate_and_update(**attributes)

    def remove(self, obj):
        if self.db[self._field_._tablename][self._field_.name]._isrefers:
            return self.db(
                self._field_._table.id == obj.id).validate_and_update(
                **{self._field_.name: None}
            )
        return self.db(self._field_._table.id == obj.id).delete()


class HasManyViaSet(RelationSet):
    _relation_method_ = 'via'
    _via_error = "Can't %s elements in has_many relations without a join table"

    @cachedprop
    def _viadata(self):
        query, rfield, model_name, rid, via, viadata = \
            super(HasManyViaSet, self)._get_query_()
        return sdict(
            query=query, rfield=rfield, model_name=model_name, rid=rid,
            via=via, data=viadata
        )

    def _get_query_(self):
        return self._viadata.query

    @property
    def _model_(self):
        return self.db[self._viadata.model_name]._model_

    def _cache_resultset(self):
        return self.select(self._viadata.rfield)

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args:
            if not kwargs:
                return self._last_resultset(refresh)
            args = [self._viadata.rfield]
        return self.select(*args, **kwargs)

    def _get_relation_fields(self):
        viadata = self._viadata.data
        self_field = self._model_._hasmany_ref_[viadata.via].field
        rel_field = viadata.field or viadata.name[:-1]
        return self_field, rel_field

    def _fields_from_scopes(self):
        viadata = self._viadata.data
        rel = self._model_._hasmany_ref_[viadata.via]
        if rel.method:
            scopes = [lambda: rel.dbset.query]
        else:
            scopes = self._relation_._extra_scopes(rel)
        return self._get_fields_from_scopes(scopes, rel.table_name)

    def create(self, **kwargs):
        raise RuntimeError('Cannot create third objects for many relations')

    def add(self, obj, **kwargs):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'add')
        nrow = self._fields_from_scopes()
        nrow.update(**kwargs)
        #: get belongs references
        self_field, rel_field = self._get_relation_fields()
        nrow[self_field] = self._viadata.rid
        nrow[rel_field] = obj.id
        #: validate and insert
        return self.db[self._viadata.via]._model_.create(nrow)

    def remove(self, obj):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'remove')
        #: get belongs references
        self_field, rel_field = self._get_relation_fields()
        #: delete
        return self.db(
            (self.db[self._viadata.via][self_field] == self._viadata.rid) &
            (self.db[self._viadata.via][rel_field] == obj.id)).delete()


class JoinedSet(Set):
    @classmethod
    def _from_set(cls, obj, jdata=[], ljdata=[], auto_select_tables=[]):
        rv = cls(
            obj.db, obj.query, obj.query.ignore_common_filters, obj._model_)
        rv._stable_ = obj._model_.tablename
        rv._jdata_ = list(jdata)
        rv._ljdata_ = list(ljdata)
        rv._auto_select_tables_ = list(auto_select_tables)
        return rv

    def _clone(self, ignore_common_filters=None, model=None, **changes):
        rv = super(JoinedSet, self)._clone(
            ignore_common_filters, model, **changes)
        rv._stable_ = self._stable_
        rv._jdata_ = self._jdata_
        rv._ljdata_ = self._ljdata_
        rv._auto_select_tables_ = self._auto_select_tables_
        return rv

    def _join_set_builder(self, obj, jdata, auto_select_tables):
        return JoinedSet._from_set(
            obj, jdata=self._jdata_ + jdata, ljdata=self._ljdata_,
            auto_select_tables=self._auto_select_tables_ + auto_select_tables)

    def _left_join_set_builder(self, jdata):
        return JoinedSet._from_set(
            self, jdata=self._jdata_, ljdata=self._ljdata_ + jdata,
            auto_select_tables=self._auto_select_tables_)

    def _iterselect_rows(self, *fields, **attributes):
        tablemap = self.db._adapter.tables(
            self.query, attributes.get('join', None),
            attributes.get('left', None), attributes.get('orderby', None),
            attributes.get('groupby', None))
        fields = self.db._adapter.expand_all(fields, tablemap)
        colnames, sql = self.db._adapter._select_wcols(
            self.query, fields, **attributes)
        return JoinIterRows(self.db, sql, fields, colnames)

    def _split_joins(self, joins):
        rv = {'belongs': [], 'one': [], 'many': []}
        for jname, jtable, rel_type in joins:
            rv[rel_type].append((jname, jtable))
        return rv['belongs'], rv['one'], rv['many']

    def _build_records_from_joined(self, rowmap, inclusions, colnames):
        for rid, many_data in inclusions.items():
            for jname, included in many_data.items():
                rowmap[rid][jname]._cached_resultset = Rows(
                    self.db, list(included.values()), [])
        return JoinRows(
            self.db, list(rowmap.values()), colnames,
            _jdata=self._jdata_ + self._ljdata_)

    def _run_select_(self, *fields, **options):
        #: build parsers
        belongs_j, one_j, many_j = self._split_joins(self._jdata_)
        belongs_l, one_l, many_l = self._split_joins(self._ljdata_)
        parsers = (
            self._build_jparsers(belongs_j, one_j, many_j) +
            self._build_lparsers(belongs_l, one_l, many_l)
        )
        #: auto add selection field for left joins
        if self._ljdata_:
            fields = list(fields)
            if not fields:
                fields = [v.ALL for v in self._auto_select_tables_]
            for join in options['left']:
                fields.append(join.first.ALL)
        #: use iterselect for performance
        rows = self._iterselect_rows(*fields, **options)
        #: rebuild rowset using nested objects
        rowmap = OrderedDict()
        inclusions = defaultdict(
            lambda: {
                jname: OrderedDict() for jname, jtable in (many_j + many_l)})
        for row in rows:
            rid = row[self._stable_].id
            rowmap[rid] = rowmap.get(rid, row[self._stable_])
            for parser in parsers:
                parser(rowmap, inclusions, row, rid)
        return self._build_records_from_joined(
            rowmap, inclusions, rows.colnames)

    def _build_jparsers(self, belongs, one, many):
        rv = []
        for jname, jtable in belongs:
            rv.append(self._jbelong_parser(jname, jtable, self.db))
        for jname, jtable in one:
            rv.append(self._jone_parser(jname, jtable))
        for jname, jtable in many:
            rv.append(self._jmany_parser(jname, jtable))
        return rv

    def _build_lparsers(self, belongs, one, many):
        rv = []
        for jname, jtable in belongs:
            rv.append(self._lbelong_parser(jname, jtable, self.db))
        for jname, jtable in one:
            rv.append(self._lone_parser(jname, jtable))
        for jname, jtable in many:
            rv.append(self._lmany_parser(jname, jtable))
        return rv

    @staticmethod
    def _jbelong_parser(fieldname, tablename, db):
        def parser(rowmap, inclusions, row, rid):
            rowmap[rid][fieldname] = JoinedIDReference._from_record(
                row[tablename], db[tablename])
        return parser

    @staticmethod
    def _jone_parser(fieldname, tablename):
        def parser(rowmap, inclusions, row, rid):
            rowmap[rid][fieldname]._cached_resultset = row[tablename]
        return parser

    @staticmethod
    def _jmany_parser(fieldname, tablename):
        def parser(rowmap, inclusions, row, rid):
            inclusions[rid][fieldname][row[tablename].id] = \
                inclusions[rid][fieldname].get(
                    row[tablename].id, row[tablename])
        return parser

    @staticmethod
    def _lbelong_parser(fieldname, tablename, db):
        def parser(rowmap, inclusions, row, rid):
            if not row[tablename].id:
                return
            rowmap[rid][fieldname] = JoinedIDReference._from_record(
                row[tablename], db[tablename])
        return parser

    @staticmethod
    def _lone_parser(fieldname, tablename):
        def parser(rowmap, inclusions, row, rid):
            if not row[tablename].id:
                return
            rowmap[rid][fieldname]._cached_resultset = row[tablename]
        return parser

    @staticmethod
    def _lmany_parser(fieldname, tablename):
        def parser(rowmap, inclusions, row, rid):
            if not row[tablename].id:
                return
            inclusions[rid][fieldname][row[tablename].id] = \
                inclusions[rid][fieldname].get(
                    row[tablename].id, row[tablename])
        return parser


class Row(_Row):
    _as_dict_types_ = tuple(
        [type(None)] + [int, float, bool, list, dict, str] +
        [datetime.datetime, datetime.date, datetime.time])

    def as_dict(self, datetime_to_str=False, custom_types=None):
        rv = {}
        for key, val in self.items():
            if isinstance(val, Row):
                val = val.as_dict()
            elif isinstance(val, _IDReference):
                val = int(val)
            elif isinstance(val, decimal.Decimal):
                val = float(val)
            elif not isinstance(val, self._as_dict_types_):
                continue
            rv[key] = val
        return rv

    def __getstate__(self):
        return self.as_dict()

    def __json__(self):
        return self.as_dict()

    def __xml__(self, key=None, quote=True):
        return xml_encode(self.as_dict(), key or 'row', quote)

    def __str__(self):
        return '<Row {}>'.format(self.as_dict())

    def __repr__(self):
        return str(self)

    def __getitem__(self, name):
        try:
            return super(Row, self).__getitem__(name)
        except KeyError:
            raise KeyError(name)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)


class Rows(_Rows):
    def __init__(
        self, db=None, records=[], colnames=[], compact=True, rawrows=None
    ):
        self.db = db
        self.records = records
        self.colnames = colnames
        self._rowkeys_ = list(self.records[0].keys()) if self.records else []
        self._getrow = self._getrow_compact_ if self.compact else self._getrow_

    @cachedprop
    def compact(self):
        if not self.records:
            return False
        return len(self._rowkeys_) == 1 and self._rowkeys_[0] != '_extra'

    @cachedprop
    def compact_tablename(self):
        if not self._rowkeys_:
            return None
        return self._rowkeys_[0]

    def _getrow_(self, i):
        return self.records[i]

    def _getrow_compact_(self, i):
        return self._getrow_(i)[self.compact_tablename]

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.db, self.records[i], self.colnames)
        return self._getrow(i)

    def column(self, column=None):
        colname = str(column) if column else self.colnames[0]
        return [r[colname] for r in self]

    def sorted(self, f, reverse=False):
        if self.compact:
            keyf = lambda r: f(r[self.compact_tablename])
        else:
            keyf = f
        return [r for r in sorted(self.records, key=keyf, reverse=reverse)]

    def sort(self, f, reverse=False):
        self.records = self.sorted(f, reverse)

    def append(self, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.append(row)

    def insert(self, position, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.insert(position, row)

    def render(self, *args, **kwargs):
        raise NotImplementedError

    def as_list(self, datetime_to_str=False, custom_types=None):
        return [item.as_dict(datetime_to_str, custom_types) for item in self]

    def as_dict(self, key='id', datetime_to_str=False, custom_types=None):
        if '.' in key:
            splitted_key = key.split('.')
            keyf = lambda row: row[splitted_key[0]][splitted_key[1]]
        else:
            keyf = lambda row: row[key]
        return {keyf(item): item for item in self.as_list()}

    def __json__(self):
        return [item.__json__() for item in self]

    def __xml__(self, key=None, quote=True):
        key = key or 'rows'
        return tag[key](*[item.__xml__(quote=quote) for item in self])

    def __str__(self):
        return str(self.records)


class JoinIterRows(_IterRows):
    def __init__(self, db, sql, fields, colnames):
        self.db = db
        self.fields = fields
        self.colnames = colnames
        self.fdata, self.tables = \
            self.db._adapter._parse_expand_colnames(fields)
        self.cursor = self.db._adapter.cursor
        self.db._adapter.execute(sql)
        self.db._adapter.lock_cursor(self.cursor)
        self._head = None
        self.last_item = None
        self.last_item_id = None
        self.blob_decode = True
        self.cacheable = False
        self.sql = sql

    def __next__(self):
        db_row = self.cursor.fetchone()
        if db_row is None:
            raise StopIteration
        return self.db._adapter._parse(
            db_row, self.fdata, self.tables, self.fields, self.colnames,
            self.blob_decode)

    def __iter__(self):
        try:
            row = next(self)
            while row is not None:
                yield row
                row = next(self)
        except StopIteration:
            self.db._adapter.close_cursor(self.cursor)
        return


class JoinRows(Rows):
    def __init__(self, *args, **kwargs):
        self._joins_ = kwargs.pop('_jdata')
        super(JoinRows, self).__init__(*args, **kwargs)

    def as_list(
        self, compact=True, storage_to_dict=True, datetime_to_str=False,
        custom_types=None
    ):
        if storage_to_dict:
            items = []
            for row in self:
                item = row.as_dict(datetime_to_str, custom_types)
                for jdata in self._joins_:
                    if not jdata[2]:
                        item[jdata[0]] = row[jdata[0]].as_list()
                items.append(item)
        else:
            items = [item for item in self]
        return items



# -*- coding: utf-8 -*-
"""
    emmett.tools.auth.models
    ------------------------

    Provides models for the authorization system.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ...ctx import current
from ...locals import now, request
from ...orm import Model, Field, before_insert, rowmethod
from ...security import uuid


class TimestampedModel(Model):
    created_at = Field.datetime(default=now, rw=False)
    updated_at = Field.datetime(default=now, update=now, rw=False)


class AuthModel(Model):
    _additional_inheritable_dict_attrs_ = [
        'form_registration_rw', 'form_profile_rw']
    auth = None

    @classmethod
    def _init_inheritable_dicts_(cls):
        for attr in cls._additional_inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name, default = attr
            else:
                attr_name, default = attr, {}
            if not isinstance(default, dict):
                raise SyntaxError(
                    "{} is not a dictionary".format(attr_name))
            setattr(cls, attr_name, default)

    @classmethod
    def _merge_inheritable_dicts_(cls, models):
        super(AuthModel, cls)._merge_inheritable_dicts_(models)
        for attr in cls._additional_inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name = attr[0]
            else:
                attr_name = attr
            attrs = {}
            for model in models:
                if not issubclass(model, AuthModel):
                    continue
                superattrs = getattr(model, attr_name)
                for k, v in superattrs.items():
                    attrs[k] = v
            for k, v in getattr(cls, attr_name).items():
                attrs[k] = v
            setattr(cls, attr_name, attrs)

    def __super_method(self, name):
        return getattr(super(AuthModel, self), '_Model__' + name)

    def _define_(self):
        self.__hide_all()
        self.__super_method('define_indexes')()
        self.__super_method('define_validation')()
        self.__super_method('define_access')()
        self.__super_method('define_defaults')()
        self.__super_method('define_updates')()
        self.__super_method('define_representation')()
        self.__super_method('define_computations')()
        self.__super_method('define_callbacks')()
        self.__super_method('define_scopes')()
        self.__super_method('define_form_utils')()
        self.__define_authform_utils()
        self.setup()

    #def __define_extra_fields(self):
    #    self.auth.settings.extra_fields['auth_user'] = self.fields

    def __hide_all(self):
        alwaysvisible = ['first_name', 'last_name', 'password', 'email']
        for field in self.table.fields:
            if field not in alwaysvisible:
                self.table[field].writable = self.table[field].readable = \
                    False

    def __base_visibility(self):
        rv = {}
        for field in self.table:
            rv[field.name] = field.readable, field.writable
        return rv

    def __define_authform_utils(self):
        settings = {
            'form_registration_rw': {'writable': [], 'readable': []},
            'form_profile_rw': {'writable': [], 'readable': []}}
        for config_dict in settings.keys():
            rw_data = self.__base_visibility()
            rw_data.update(**self.fields_rw)
            rw_data.update(**getattr(self, config_dict))
            for key, value in rw_data.items():
                if isinstance(value, (tuple, list)):
                    readable, writable = value
                else:
                    readable = writable = value
                if readable:
                    settings[config_dict]['readable'].append(key)
                if writable:
                    settings[config_dict]['writable'].append(key)
        setattr(self, '_merged_form_rw_', {
            'registration': settings['form_registration_rw'],
            'profile': settings['form_profile_rw']})


class AuthUserBasic(AuthModel, TimestampedModel):
    tablename = "auth_users"
    format = '%(email)s (%(id)s)'
    #: injected by Auth
    #  has_many(
    #      {'auth_memberships': 'AuthMembership'},
    #      {'auth_events': 'AuthEvent'},
    #      {'auth_groups': {'via': 'auth_memberships'}},
    #      {'auth_permissions': {'via': 'auth_groups'}},
    #  )

    email = Field(length=255, unique=True)
    password = Field.password(length=512)
    registration_key = Field(length=512, rw=False, default='')
    reset_password_key = Field(length=512, rw=False, default='')
    registration_id = Field(length=512, rw=False, default='')

    form_labels = {
        'email': 'E-mail',
        'password': 'Password'
    }

    form_profile_rw = {
        'email': (True, False),
        'password': False
    }

    @before_insert
    def set_registration_key(self, fields):
        if self.auth.config.registration_verification and not \
                fields.get('registration_key'):
            fields['registration_key'] = uuid()
        elif self.auth.config.registration_approval:
            fields['registration_key'] = 'pending'

    @rowmethod('disable')
    def _set_disabled(self, row):
        return row.update_record(registration_key='disabled')

    @rowmethod('block')
    def _set_blocked(self, row):
        return row.update_record(registration_key='blocked')

    @rowmethod('allow')
    def _set_allowed(self, row):
        return row.update_record(registration_key='')


class AuthUser(AuthUserBasic):
    format = '%(first_name)s %(last_name)s (%(id)s)'

    first_name = Field(length=128, notnull=True)
    last_name = Field(length=128, notnull=True)

    form_labels = {
        'first_name': 'First name',
        'last_name': 'Last name',
    }


class AuthGroup(TimestampedModel):
    format = '%(role)s (%(id)s)'
    #: injected by Auth
    #  has_many(
    #      {'auth_memberships': 'AuthMembership'},
    #      {'auth_permissions': 'AuthPermission'},
    #      {'users': {'via': 'memberships'}}
    #  )

    role = Field(length=255, default='', unique=True)
    description = Field.text()

    form_labels = {
        'role': 'Role',
        'description': 'Description'
    }


class AuthMembership(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'user': 'AuthUser'}, {'auth_group': 'AuthGroup'})
    pass


class AuthPermission(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'auth_group': 'AuthGroup'})

    name = Field(length=512, default='default', notnull=True)
    table_name = Field(length=512)
    record_id = Field.int(default=0)

    validation = {
        'record_id': {'in': {'range': (0, 10**9)}}
    }

    form_labels = {
        'name': 'Name',
        'table_name': 'Object or table name',
        'record_id': 'Record ID'
    }


class AuthEvent(TimestampedModel):
    #: injected by Auth
    #  belongs_to({'user': 'AuthUser'})

    client_ip = Field()
    origin = Field(length=512, notnull=True)
    description = Field.text(notnull=True)

    default_values = {
        'client_ip': lambda:
            request.client if hasattr(current, 'request') else 'unavailable',
        'origin': 'auth',
        'description': ''
    }

    #: labels injected by Auth
    form_labels = {
        'client_ip': 'Client IP',
        'origin': 'Origin',
        'description': 'Description'
    }



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.response
    ------------------------

    Provides response wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from http.cookies import Morsel, SimpleCookie
from typing import Any

from ..datastructures import sdict
from ..helpers import get_flashed_messages
from ..html import htmlescape
from ..utils import cachedprop
from . import Wrapper
from .helpers import ResponseHeaders

# Workaround for adding samesite support to pre 3.8 python
Morsel._reserved["samesite"] = "SameSite"  # type: ignore  # noqa


class Response(Wrapper):
    __slots__ = ('status', 'headers', 'cookies')

    def __init__(self):
        self.status = 200
        self.headers = ResponseHeaders({'content-type': 'text/plain'})
        self.cookies = SimpleCookie()

    @cachedprop
    def meta(self) -> sdict[str, Any]:
        return sdict()

    @cachedprop
    def meta_prop(self) -> sdict[str, Any]:
        return sdict()

    @property
    def content_type(self) -> str:
        return self.headers['content-type']

    @content_type.setter
    def content_type(self, value: str):
        self.headers['content-type'] = value

    def alerts(self, **kwargs):
        return get_flashed_messages(**kwargs)

    def _meta_tmpl(self):
        return [
            (key, htmlescape(val)) for key, val in self.meta.items()
        ]

    def _meta_tmpl_prop(self):
        return [
            (key, htmlescape(val)) for key, val in self.meta_prop.items()
        ]



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.helpers
    -----------------------

    Provides wrappers helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union
)

from .._internal import loop_copyfileobj


class Headers(Mapping[str, str]):
    __slots__ = ["_data"]

    def __init__(self, scope: Dict[str, Any]):
        self._data: Dict[bytes, bytes] = {
            key: val for key, val in scope["headers"]
        }

    __hash__ = None  # type: ignore

    def __getitem__(self, key: str) -> str:
        return self._data[key.lower().encode("latin-1")].decode("latin-1")

    def __contains__(self, key: str) -> bool:  # type: ignore
        return key.lower().encode("latin-1") in self._data

    def __iter__(self) -> Iterator[str]:
        for key in self._data.keys():
            yield key.decode("latin-1")

    def __len__(self) -> int:
        return len(self._data)

    def get(
        self,
        key: str,
        default: Optional[Any] = None,
        cast: Optional[Callable[[Any], Any]] = None
    ) -> Any:
        rv = self._data.get(key.lower().encode("latin-1"))
        rv = rv.decode() if rv is not None else default  # type: ignore
        if cast is None:
            return rv
        try:
            return cast(rv)
        except ValueError:
            return default

    def items(self) -> Iterator[Tuple[str, str]]:  # type: ignore
        for key, value in self._data.items():
            yield key.decode("latin-1"), value.decode("latin-1")

    def keys(self) -> Iterator[str]:  # type: ignore
        for key in self._data.keys():
            yield key.decode("latin-1")

    def values(self) -> Iterator[str]:  # type: ignore
        for value in self._data.values():
            yield value.decode("latin-1")


class ResponseHeaders(MutableMapping[str, str]):
    __slots__ = ['_data']

    def __init__(self, data: Optional[Dict[str, str]] = None):
        self._data = data or {}

    __hash__ = None  # type: ignore

    def __getitem__(self, key: str) -> str:
        return self._data[key.lower()]

    def __setitem__(self, key: str, value: str):
        self._data[key.lower()] = value

    def __delitem__(self, key: str):
        del self._data[key.lower()]

    def __contains__(self, key: str) -> bool:  # type: ignore
        return key.lower() in self._data

    def __iter__(self) -> Iterator[str]:
        for key in self._data.keys():
            yield key

    def __len__(self) -> int:
        return len(self._data)

    def items(self) -> Iterator[Tuple[str, str]]:  # type: ignore
        for key, value in self._data.items():
            yield key, value

    def keys(self) -> Iterator[str]:  # type: ignore
        for key in self._data.keys():
            yield key

    def values(self) -> Iterator[str]:  # type: ignore
        for value in self._data.values():
            yield value

    def update(self, data: Dict[str, str]):  # type: ignore
        self._data.update(data)


class FileStorage:
    __slots__ = ('stream', 'filename', 'name', 'headers', 'content_type')

    def __init__(
        self,
        stream: BinaryIO,
        filename: str,
        name: str = None,
        content_type: str = None,
        headers: Dict = None
    ):
        self.stream = stream
        self.filename = filename
        self.name = name
        self.headers = headers or {}
        self.content_type = content_type or self.headers.get('content-type')

    @property
    def content_length(self) -> int:
        return int(self.headers.get('content-length', 0))

    async def save(
        self,
        destination: Union[BinaryIO, str],
        buffer_size: int = 16384
    ):
        close_destination = False
        if isinstance(destination, str):
            destination = open(destination, 'wb')
            close_destination = True
        try:
            await loop_copyfileobj(self.stream, destination, buffer_size)
        finally:
            if close_destination:
                destination.close()

    def __iter__(self) -> Iterable[bytes]:
        return iter(self.stream)

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__}: '
            f'{self.filename} ({self.content_type})')


class RequestCancelled(Exception):
    pass



# -*- coding: utf-8 -*-
"""
    emmett.tools.auth.apis
    ----------------------

    Provides the interface for the auth system.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pydal.helpers.classes import Reference as _RecordReference

from ...cache import RouteCacheRule
from ...datastructures import sdict
from ...locals import now, session
from ...orm.objects import Row
from ...orm.models import Model
from ...pipeline import Pipe, Injector
from .ext import AuthExtension
from .exposer import AuthModule


class Auth:
    def __init__(self, app, db, user_model=None):
        self.ext = app.use_extension(AuthExtension)
        self.ext.bind_auth(self)
        self.ext.use_database(db, user_model)
        self.ext.init_forms()
        self.pipe = AuthPipe(self)

    def module(
        self,
        import_name: str,
        name: str = 'auth',
        template_folder: str = 'auth',
        template_path: Optional[str] = None,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = 'auth',
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        injectors: Optional[List[Injector]] = None,
        module_class: Type[AuthModule] = AuthModule,
        **kwargs: Any
    ) -> AuthModule:
        return module_class.from_app(
            self.ext.app,
            import_name,
            name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline or [],
            injectors=injectors or [],
            opts=kwargs
        )

    @property
    def models(self) -> Dict[str, Model]:
        return self.ext.config.models

    def group_for_role(self, role: str) -> Row:
        return self.models['group'].get(role=role)

    #: context
    @property
    def session(self) -> sdict:
        return session.auth

    @property
    def user(self) -> Optional[Row]:
        try:
            rv = self.session.user
        except Exception:
            rv = None
        return rv

    #: helpers
    def is_logged(self) -> bool:
        return True if self.user else False

    def has_membership(
        self,
        group: Optional[Union[str, int, Row]] = None,
        user: Optional[Union[Row, int]] = None,
        role: Optional[str] = None
    ) -> bool:
        rv = False
        if not group and role:
            group = self.group_for_role(role)
        if isinstance(group, str):
            group = self.group_for_role(group)
        if not user and self.user:
            user = self.user.id
        if group and user:
            if self.models['membership'].where(
                lambda m:
                    (m.table[self.ext.relation_names['user']] == user) &
                    (m.table[self.ext.relation_names['group']] == group)
            ).count():
                rv = True
        return rv

    def has_permission(
        self,
        name: str = 'any',
        table_name: Optional[str] = None,
        record_id: Optional[int] = None,
        user: Optional[Union[int, Row]] = None,
        group: Optional[Union[str, int, Row]] = None
    ) -> bool:
        permission = self.models['permission']
        parent = None
        query = (permission.name == name)
        if table_name:
            query = query & (permission.table_name == table_name)
        if record_id:
            query = query & (permission.record_id == record_id)
        if not user and self.user:
            user = self.user.id
        if not user and not group:
            return False
        if user is not None:
            parent = self.models['user'].get(id=user)
        elif group is not None:
            if isinstance(group, str):
                group = self.group_for_role(group)
            parent = self.models['group'].get(id=group)
        if not parent:
            return False
        return (
            parent[self.ext.relation_names['permission'] + 's'].where(
                query
            ).count() > 0
        )

    #: operations
    def create_group(self, role: str, description: str = '') -> _RecordReference:
        res = self.models['group'].create(
            role=role, description=description
        )
        return res.id

    def delete_group(self, group: Union[str, int, Row]) -> int:
        if isinstance(group, str):
            group = self.group_for_role(group)
        return self.ext.db(self.models['group'].id == group).delete()

    def add_membership(
        self,
        group: Union[str, int, Row],
        user: Optional[Row] = None
    ) -> _RecordReference:
        if isinstance(group, int):
            group = self.models['group'].get(group)
        elif isinstance(group, str):
            group = self.group_for_role(group)
        if not user and self.user:
            user = self.user.id
        res = getattr(
            group, self.ext.relation_names['user'] + 's'
        ).add(user)
        return res.id

    def remove_membership(
        self,
        group: Union[str, int, Row],
        user: Optional[Row] = None
    ):
        if isinstance(group, int):
            group = self.models['group'].get(group)
        elif isinstance(group, str):
            group = self.group_for_role(group)
        if not user and self.user:
            user = self.user.id
        return getattr(
            group, self.ext.relation_names['user'] + 's'
        ).remove(user)

    def login(self, email: str, password: str):
        user = self.models['user'].get(email=email)
        if user and user.get('password', False):
            password = self.models['user'].password.validate(password)[0]
            if not user.registration_key and password == user.password:
                self.ext.login_user(user)
                return user
        return None

    def change_user_status(self, user: Union[int, Row], status: str) -> int:
        return self.ext.db(self.models['user'].id == user).update(
            registration_key=status
        )

    def disable_user(self, user: Union[int, Row]) -> int:
        return self.change_user_status(user, 'disabled')

    def block_user(self, user: Union[int, Row]) -> int:
        return self.change_user_status(user, 'blocked')

    def allow_user(self, user: Union[int, Row]) -> int:
        return self.change_user_status(user, '')

    #: emails decorators
    def registration_mail(
        self,
        f: Callable[[Row, Dict[str, Any]], bool]
    ) -> Callable[[Row, Dict[str, Any]], bool]:
        self.ext.mails['registration'] = f
        return f

    def reset_password_mail(
        self,
        f: Callable[[Row, Dict[str, Any]], bool]
    ) -> Callable[[Row, Dict[str, Any]], bool]:
        self.ext.mails['reset_password'] = f
        return f


class AuthPipe(Pipe):
    def __init__(self, auth):
        #: the Auth() instance
        self.auth = auth

    def session_open(self):
        # check auth session is valid
        authsess = self.auth.session
        if not authsess:
            return
        #: check session has needed data
        if not authsess.last_visit or not authsess.last_dbcheck:
            del session.auth
            return
        #: is session expired?
        visit_dt = now().as_naive_datetime()
        if (
            authsess.last_visit + timedelta(seconds=authsess.expiration) <
            visit_dt
        ):
            del session.auth
        #: does session need re-sync with db?
        elif authsess.last_dbcheck + timedelta(seconds=300) < visit_dt:
            if self.auth.user:
                #: is user still valid?
                dbrow = self.auth.models['user'].get(self.auth.user.id)
                if dbrow and not dbrow.registration_key:
                    self.auth.ext.login_user(dbrow, authsess.remember)
                else:
                    del session.auth
        else:
            #: set last_visit if make sense
            if (
                (visit_dt - authsess.last_visit).seconds >
                min(authsess.expiration / 10, 600)
            ):
                authsess.last_visit = visit_dt

    def session_close(self):
        # set correct session expiration if requested by user
        if self.auth.session and self.auth.session.remember:
            session._expires_after(self.auth.session.expiration)

    async def pipe(self, next_pipe, **kwargs):
        self.session_open()
        return await next_pipe(**kwargs)

    async def on_pipe_success(self):
        self.session_close()

    async def on_pipe_failure(self):
        self.session_close()



# -*- coding: utf-8 -*-
"""
    emmett.orm.wrappers
    -------------------

    Provides ORM wrappers utilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from .helpers import RelationBuilder
from .objects import HasOneSet, HasManySet, HasManyViaSet


class Wrapper(object):
    def __init__(self, ref):
        self.__name__ = ref.name
        self.ref = ref


class HasOneWrap(Wrapper):
    def __call__(self, model, row):
        return HasOneSet(model.db, RelationBuilder(self.ref, model), row)


class HasManyWrap(Wrapper):
    def __call__(self, model, row):
        return HasManySet(model.db, RelationBuilder(self.ref, model), row)


class HasManyViaWrap(Wrapper):
    def __call__(self, model, row):
        return HasManyViaSet(model.db, RelationBuilder(self.ref, model), row)



# -*- coding: utf-8 -*-
"""
    tests.utils
    -----------

    Test Emmett utils engine
"""

import pytest

from emmett.datastructures import sdict
from emmett.utils import (
    cachedprop, _cached_prop_sync, _cached_prop_loop,
    dict_to_sdict, is_valid_ip_address)


class Class:
    def __init__(self):
        self.calls = 0

    @cachedprop
    def prop(self):
        self.calls += 1
        return 'test_cachedprop_sync'

    @cachedprop
    async def prop_loop(self):
        self.calls += 1
        return 'test_cachedprop_loop'


def test_cachedprop_sync():
    assert isinstance(Class.prop, _cached_prop_sync)
    obj = Class()
    assert obj.calls == 0
    assert obj.prop == 'test_cachedprop_sync'
    assert obj.calls == 1
    assert obj.prop == 'test_cachedprop_sync'
    assert obj.calls == 1


@pytest.mark.asyncio
async def test_cachedprop_loop():
    assert isinstance(Class.prop_loop, _cached_prop_loop)
    obj = Class()
    assert obj.calls == 0
    assert (await obj.prop_loop) == 'test_cachedprop_loop'
    assert obj.calls == 1
    assert (await obj.prop_loop) == 'test_cachedprop_loop'
    assert obj.calls == 1


def test_is_valid_ip_address():
    result_localhost = is_valid_ip_address('127.0.0.1')
    assert result_localhost is True

    result_unknown = is_valid_ip_address('unknown')
    assert result_unknown is False

    result_ipv4_valid = is_valid_ip_address('::ffff:192.168.0.1')
    assert result_ipv4_valid is True

    result_ipv4_valid = is_valid_ip_address('192.168.256.1')
    assert result_ipv4_valid is False

    result_ipv6_valid = is_valid_ip_address('fd40:363d:ee85::')
    assert result_ipv6_valid is True

    result_ipv6_valid = is_valid_ip_address('fd40:363d:ee85::1::')
    assert result_ipv6_valid is False


def test_dict_to_sdict():
    result_sdict = dict_to_sdict({'test': 'dict'})
    assert isinstance(result_sdict, sdict)
    assert result_sdict.test == 'dict'

    result_number = dict_to_sdict(1)
    assert not isinstance(result_number, sdict)
    assert result_number == 1



# -*- coding: utf-8 -*-
"""
    emmett.orm.adapters
    -------------------

    Provides ORM adapters facilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from functools import wraps
from pydal.adapters import adapters
from pydal.adapters.mssql import (
    MSSQL1,
    MSSQL3,
    MSSQL4,
    MSSQL1N,
    MSSQL3N,
    MSSQL4N
)
from pydal.adapters.postgres import (
    Postgre,
    PostgrePsyco,
    PostgrePG8000,
    PostgreNew,
    PostgrePsycoNew,
    PostgrePG8000New
)

from .engines import adapters
from .objects import Field


adapters._registry_.update({
    'mssql': MSSQL4,
    'mssql2': MSSQL1,
    'mssql3': MSSQL3,
    'mssqln': MSSQL4N,
    'mssqln2': MSSQL1N,
    'mssqln3': MSSQL3N,
    'postgres2': PostgreNew,
    'postgres2:psycopg2': PostgrePsycoNew,
    'postgres2:pg8000': PostgrePG8000New,
    'postgres3': Postgre,
    'postgres3:psycopg2': PostgrePsyco,
    'postgres3:pg8000': PostgrePG8000
})


def _wrap_on_obj(f, adapter):
    @wraps(f)
    def wrapped(*args, **kwargs):
        return f(adapter, *args, **kwargs)
    return wrapped


def patch_adapter(adapter):
    adapter.parse = _wrap_on_obj(parse, adapter)
    adapter._parse_expand_colnames = _wrap_on_obj(
        _parse_expand_colnames, adapter)
    adapter._parse = _wrap_on_obj(_parse, adapter)
    patch_dialect(adapter.dialect)


def patch_dialect(dialect):
    _create_table_map = {
        'mysql': _create_table_mysql,
        'firebird': _create_table_firebird
    }
    dialect.create_table = _wrap_on_obj(
        _create_table_map.get(dialect.adapter.dbengine, _create_table),
        dialect)


def parse(adapter, rows, fields, colnames, blob_decode=True, cacheable=False):
    fdata, tables = _parse_expand_colnames(adapter, fields)
    new_rows = [
        _parse(adapter, row, fdata, tables, fields, colnames, blob_decode)
        for row in rows
    ]
    rowsobj = adapter.db.Rows(adapter.db, new_rows, colnames, rawrows=rows)
    return rowsobj


def _parse_expand_colnames(adapter, fieldlist):
    rv, tables = [], {}
    for field in fieldlist:
        if not isinstance(field, Field):
            rv.append(None)
            continue
        table = field.table
        tablename, fieldname = table._tablename, field.name
        ft = field.type
        fit = field._itype
        rv.append((tablename, fieldname, table, field, ft, fit))
        tables[tablename] = table
    return rv, tables


def _parse(adapter, row, fdata, tables, fields, colnames, blob_decode):
    new_row = _build_newrow_wtables(adapter, tables)
    extras = adapter.db.Row()
    #: let's loop over columns
    for (idx, colname) in enumerate(colnames):
        value = row[idx]
        fd = fdata[idx]
        tablename = None
        #: do we have a real column?
        if fd:
            (tablename, fieldname, table, field, ft, fit) = fd
            colset = new_row[tablename]
            #: parse value
            value = adapter.parse_value(value, fit, ft, blob_decode)
            if field.filter_out:
                value = field.filter_out(value)
            colset[fieldname] = value
        #: otherwise we set the value in extras
        else:
            value = adapter.parse_value(
                value, fields[idx]._itype, fields[idx].type, blob_decode)
            extras[colname] = value
            new_column_name = adapter._regex_select_as_parser(colname)
            if new_column_name is not None:
                column_name = new_column_name.groups(0)
                new_row[column_name[0]] = value
    #: add extras if needed (eg. operations results)
    if extras:
        new_row['_extra'] = extras
    return new_row


def _build_newrow_wtables(adapter, tables):
    rv = adapter.db.Row()
    for name, table in tables.items():
        rv[name] = table._model_._rowclass_()
    return rv


def _create_table(dialect, tablename, fields):
    return [
        "CREATE TABLE %s(\n    %s\n);" % (dialect.quote(tablename), fields)]


def _create_table_mysql(dialect, tablename, fields):
    return ["CREATE TABLE %s(\n    %s\n) ENGINE=%s CHARACTER SET utf8;" % (
        dialect.quote(tablename), fields,
        dialect.adapter.adapter_args.get('engine', 'InnoDB'))]


def _create_table_firebird(dialect, tablename, fields):
    rv = _create_table(dialect, tablename, fields)
    sequence_name = dialect.sequence_name(tablename)
    trigger_name = dialect.trigger_name(tablename)
    trigger_sql = (
        'create trigger %s for %s active before insert position 0 as\n'
        'begin\n'
        'if(new."id" is null) then\n'
        'begin\n'
        'new."id" = gen_id(%s, 1);\n'
        'end\n'
        'end;')
    rv.extend([
        'create generator %s;' % sequence_name,
        'set generator %s to 0;' % sequence_name,
        trigger_sql % (trigger_name, dialect.quote(tablename), sequence_name)
    ])
    return rv


def _initialize(adapter, *args, **kwargs):
    adapter._find_work_folder()
    adapter._connection_manager.configure(
        max_connections=adapter.db._pool_size,
        connect_timeout=adapter.db._connect_timeout,
        stale_timeout=adapter.db._keep_alive_timeout)


def _begin(adapter):
    pass


def _in_transaction(adapter):
    return bool(adapter._connection_manager.state.transactions)


def _push_transaction(adapter, transaction):
    adapter._connection_manager.state.transactions.append(transaction)


def _pop_transaction(adapter):
    adapter._connection_manager.state.transactions.pop()


def _transaction_depth(adapter):
    return len(adapter._connection_manager.state.transactions)


def _top_transaction(adapter):
    if adapter._connection_manager.state.transactions:
        return adapter._connection_manager.state.transactions[-1]



# -*- coding: utf-8 -*-
"""
    emmett.debug
    ------------

    Provides debugging utilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import inspect
import os
import sys
import traceback

from renoir import Renoir

from .utils import cachedprop


class Traceback:
    """Wraps a traceback."""

    def __init__(self, app, exc_type, exc_value, tb):
        self.app = app
        self.exc_type = exc_type
        self.exc_value = exc_value
        if not isinstance(exc_type, str):
            exception_type = exc_type.__name__
            if exc_type.__module__ not in (
                '__builtin__', 'builtins', 'exceptions'
            ):
                exception_type = exc_type.__module__ + '.' + exception_type
        else:
            exception_type = exc_type
        self.exception_type = exception_type

        self.frames = []
        while tb:
            self.frames.append(Frame(self.app, exc_type, exc_value, tb))
            tb = tb.tb_next

    @property
    def exception(self):
        """String representation of the exception."""
        buf = traceback.format_exception_only(self.exc_type, self.exc_value)
        return ''.join(buf).strip()

    def generate_plaintext_traceback(self):
        """Like the plaintext attribute but returns a generator"""
        yield u'Traceback (most recent call last):'
        for frame in self.frames:
            yield u'  File "%s", line %s, in %s' % (
                frame.filename,
                frame.lineno,
                frame.function_name
            )
            yield u'    ' + frame.current_line.strip()
        yield self.exception

    def generate_plain_tb_app(self):
        yield u'Traceback (most recent call last):'
        for frame in self.frames:
            if frame.is_in_app:
                yield u'  File "%s", line %s, in %s' % (
                    frame.filename,
                    frame.lineno,
                    frame.function_name
                )
                yield u'    ' + frame.current_line.strip()
        yield self.exception

    @property
    def full_tb(self):
        return u'\n'.join(self.generate_plaintext_traceback())

    @property
    def app_tb(self):
        return u'\n'.join(self.generate_plain_tb_app())


class Frame:
    """A single frame in a traceback."""

    def __init__(self, app, exc_type, exc_value, tb):
        self.app = app
        self.lineno = tb.tb_lineno
        self.function_name = tb.tb_frame.f_code.co_name
        self.locals = tb.tb_frame.f_locals
        self.globals = tb.tb_frame.f_globals

        fn = inspect.getsourcefile(tb) or inspect.getfile(tb)
        if fn[-4:] in ('.pyo', '.pyc'):
            fn = fn[:-1]
        # if it's a file on the file system resolve the real filename.
        if os.path.isfile(fn):
            fn = os.path.realpath(fn)
        self.filename = fn
        self.module = self.globals.get('__name__')
        self.code = tb.tb_frame.f_code

    @property
    def is_in_fw(self):
        fw_path = os.path.dirname(__file__)
        return self.filename[0:len(fw_path)] == fw_path

    @property
    def is_in_app(self):
        return self.filename[0:len(self.app.root_path)] == self.app.root_path

    @property
    def rendered_filename(self):
        if self.is_in_app:
            return self.filename[len(self.app.root_path) + 1:]
        if self.is_in_fw:
            return ''.join([
                "emmett.",
                self.filename[
                    len(os.path.dirname(__file__)) + 1:
                ].replace("/", ".").split(".py")[0]
            ])
        return self.filename

    @cachedprop
    def sourcelines(self):
        try:
            with open(self.filename, 'rb') as file:
                source = file.read().decode('utf8')
        except IOError:
            source = '<unavailable>'
        return source.splitlines()

    @property
    def sourceblock(self):
        lmax = self.lineno + 4
        return u'\n'.join(self.sourcelines[self.first_line_no - 1:lmax])

    @property
    def first_line_no(self):
        number = self.lineno > 5 and (self.lineno - 5) or 1
        if number > len(self.sourcelines):
            number = 1
        while not self.sourcelines[number - 1]:
            number += 1
            if number > len(self.sourcelines):
                break
        return number

    @property
    def current_line(self):
        try:
            return self.sourcelines[self.lineno - 1]
        except IndexError:
            return u''

    @cachedprop
    def render_locals(self):
        rv = dict()
        for k, v in self.locals.items():
            try:
                rv[k] = str(v)
            except Exception:
                rv[k] = '<unavailable>'
        return rv


debug_templater = Renoir(
    path=os.path.join(os.path.dirname(__file__), 'assets', 'debug')
)


def smart_traceback(app):
    exc_type, exc_value, tb = sys.exc_info()
    return Traceback(app, exc_type, exc_value, tb)


def debug_handler(tb):
    return debug_templater.render('view.html', {'tb': tb})



# -*- coding: utf-8 -*-
"""
    emmett.utils
    ------------

    Provides some utilities for Emmett.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio
import re
import socket

from datetime import datetime, date, time
from typing import Any, Callable, Generic, Optional, Union, overload

import pendulum

from pendulum.parsing import _parse as _pendulum_parse

from .datastructures import sdict
from .typing import T


class _cached_prop(Generic[T]):
    def __init__(
        self,
        fget: Callable[..., T],
        name: str,
        doc: Optional[str] = None
    ):
        self.fget = fget
        self.__doc__ = doc
        self.__name__ = name

    def __get__(self, obj: Optional[object], cls: Any) -> T:
        raise NotImplementedError


def cachedprop(
    fget: Callable[..., T],
    doc: Optional[str] = None,
    name: Optional[str] = None
) -> _cached_prop[T]:
    doc = doc or fget.__doc__
    name = name or fget.__name__
    if asyncio.iscoroutinefunction(fget):
        return _cached_prop_loop[T](fget, name, doc)
    return _cached_prop_sync[T](fget, name, doc)



class _cached_prop_sync(_cached_prop[T]):
    @overload
    def __get__(self, obj: None, cls: Any) -> _cached_prop_sync:
        ...

    @overload
    def __get__(self, obj: object, cls: Any) -> T:
        ...

    def __get__(self, obj: Optional[object], cls: Any) -> Union[_cached_prop_sync, T]:
        if obj is None:
            return self
        obj.__dict__[self.__name__] = rv = self.fget(obj)
        return rv


class _cached_awaitable_coro(Generic[T]):
    slots = ['coro_f', 'obj', '_result', '_awaitable']

    def __init__(self, coro_f: Callable[..., T], obj: object):
        self.coro_f = coro_f
        self.obj = obj
        self._awaitable = self.__fetcher

    async def __fetcher(self) -> T:
        self._result = rv = await self.coro_f(self.obj)  # type: ignore
        self._awaitable = self.__cached
        return rv

    async def __cached(self) -> T:
        return self._result

    def __await__(self):
        return self._awaitable().__await__()


class _cached_prop_loop(_cached_prop[T]):
    @overload
    def __get__(self, obj: None, cls: Any) -> _cached_prop_loop:
        ...

    @overload
    def __get__(self, obj: object, cls: Any) -> T:
        ...

    def __get__(self, obj: Optional[object], cls: Any) -> Union[_cached_prop_loop, T]:
        if obj is None:
            return self
        obj.__dict__[self.__name__] = rv = _cached_awaitable_coro[T](
            self.fget, obj
        )
        return rv  # type: ignore


_pendulum_parsing_opts = {
    'day_first': False,
    'year_first': True,
    'strict': True,
    'exact': False,
    'now': None
}


def _pendulum_normalize(obj):
    if isinstance(obj, time):
        now = datetime.utcnow()
        obj = datetime(
            now.year, now.month, now.day,
            obj.hour, obj.minute, obj.second, obj.microsecond
        )
    elif isinstance(obj, date) and not isinstance(obj, datetime):
        obj = datetime(obj.year, obj.month, obj.day)
    return obj


def parse_datetime(text):
    parsed = _pendulum_normalize(
        _pendulum_parse(text, **_pendulum_parsing_opts))
    return pendulum.datetime(
        parsed.year, parsed.month, parsed.day,
        parsed.hour, parsed.minute, parsed.second, parsed.microsecond,
        tz=parsed.tzinfo or pendulum.UTC
    )


_re_ipv4 = re.compile(r'(\d+)\.(\d+)\.(\d+)\.(\d+)')


def is_valid_ip_address(address):
    # deal with special cases
    if address.lower() in [
        '127.0.0.1', 'localhost', '::1', '::ffff:127.0.0.1'
    ]:
        return True
    elif address.lower() in ('unknown', ''):
        return False
    elif address.count('.') == 3:  # assume IPv4
        if address.startswith('::ffff:'):
            address = address[7:]
        if hasattr(socket, 'inet_aton'):  # try validate using the OS
            try:
                socket.inet_aton(address)
                return True
            except socket.error:  # invalid address
                return False
        else:  # try validate using Regex
            match = _re_ipv4.match(address)
            if match and all(
                0 <= int(match.group(i)) < 256 for i in (1, 2, 3, 4)
            ):
                return True
            return False
    elif hasattr(socket, 'inet_pton'):  # assume IPv6, try using the OS
        try:
            socket.inet_pton(socket.AF_INET6, address)
            return True
        except socket.error:  # invalid address
            return False
    else:  # do not know what to do? assume it is a valid address
        return True


def read_file(filename, mode='r'):
    # returns content from filename, making sure to close the file on exit.
    f = open(filename, mode)
    try:
        return f.read()
    finally:
        f.close()


def write_file(filename, value, mode='w'):
    # writes <value> to filename, making sure to close the file on exit.
    f = open(filename, mode)
    try:
        return f.write(value)
    finally:
        f.close()


def dict_to_sdict(obj):
    #: convert dict and nested dicts to sdict
    if isinstance(obj, dict) and not isinstance(obj, sdict):
        for k in obj:
            obj[k] = dict_to_sdict(obj[k])
        return sdict(obj)
    return obj



# -*- coding: utf-8 -*-
"""
    emmett.orm.apis
    ---------------

    Provides ORM apis.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from collections import OrderedDict
from .helpers import Reference, Callback


class belongs_to(Reference):
    _references_ = OrderedDict()

    @property
    def refobj(self):
        return belongs_to._references_


class refers_to(Reference):
    _references_ = OrderedDict()

    @property
    def refobj(self):
        return refers_to._references_


class has_one(Reference):
    _references_ = OrderedDict()

    @property
    def refobj(self):
        return has_one._references_


class has_many(Reference):
    _references_ = OrderedDict()

    @property
    def refobj(self):
        return has_many._references_


class compute(object):
    _inst_count_ = 0

    def __init__(self, field_name):
        self.field_name = field_name
        self._inst_count_ = compute._inst_count_
        compute._inst_count_ += 1

    def __call__(self, f):
        self.f = f
        return self


class rowattr(object):
    _inst_count_ = 0

    def __init__(self, field_name):
        self.field_name = field_name
        self._inst_count_ = rowattr._inst_count_
        rowattr._inst_count_ += 1

    def __call__(self, f):
        self.f = f
        return self


class rowmethod(rowattr):
    pass


def before_insert(f):
    return Callback(f, '_before_insert')


def after_insert(f):
    return Callback(f, '_after_insert')


def before_update(f):
    return Callback(f, '_before_update')


def after_update(f):
    return Callback(f, '_after_update')


def before_delete(f):
    return Callback(f, '_before_delete')


def after_delete(f):
    return Callback(f, '_after_delete')


class scope(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, f):
        self.f = f
        return self



# -*- coding: utf-8 -*-
"""
    emmett.orm.helpers
    ------------------

    Provides ORM helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import re
import time

from functools import wraps
from pydal._globals import THREAD_LOCAL
from pydal.helpers.classes import Reference as _IDReference, ExecutionHandler
from pydal.objects import Field as _Field

from ..datastructures import sdict
from ..utils import cachedprop


class Reference(object):
    def __init__(self, *args, **params):
        self.reference = [arg for arg in args]
        self.params = params
        self.refobj[id(self)] = self

    def __call__(self, func):
        if self.__class__.__name__ not in ['has_one', 'has_many']:
            raise SyntaxError(
                '%s cannot be used as a decorator' % self.__class__.__name__)
        if not callable(func):
            raise SyntaxError('Argument must be callable')
        if self.reference:
            raise SyntaxError(
                "When using %s as decorator, you must use the 'field' option" %
                self.__class__.__name__)
        new_reference = {func.__name__: {'method': func}}
        field = self.params.get('field')
        if field:
            new_reference[func.__name__]['field'] = field
        cast = self.params.get('cast')
        if cast:
            new_reference[func.__name__]['cast'] = cast
        self.reference = [new_reference]
        return self

    @property
    def refobj(self):
        return {}


class ReferenceData(sdict):
    def __init__(self, model_class, **kwargs):
        self.model_class = model_class
        super(ReferenceData, self).__init__(**kwargs)

    @cachedprop
    def dbset(self):
        if self.method:
            return self.method(self.model_class)
        return self.model_class.db

    @cachedprop
    def model_instance(self):
        if self.method:
            return self.dbset._model_._instance_()
        return self.dbset[self.model]._model_

    @property
    def table(self):
        return self.model_instance.table

    @property
    def table_name(self):
        return self.model_instance.tablename

    @property
    def field_instance(self):
        return self.table[self.field]


class RelationBuilder(object):
    def __init__(self, ref, model_instance):
        self.ref = ref
        self.model = model_instance

    def _make_refid(self, row):
        return row.id if row is not None else self.model.id

    def _extra_scopes(self, ref, model_instance=None):
        model_instance = model_instance or ref.model_instance
        rv = []
        if ref.scope is not None:
            scope_m = model_instance._scopes_[ref.scope].f
            rv.append(lambda f=scope_m, m=model_instance: f(m))
        if ref.where is not None:
            rv.append(lambda f=ref.where, m=model_instance: f(m))
        return rv

    def _patch_query_with_scopes(self, ref, query):
        for scope in self._extra_scopes(ref):
            query = query & scope()
        return query

    def _patch_query_with_scopes_on(self, ref, query, model_name):
        model = self.model.db[model_name]._model_
        for scope in self._extra_scopes(ref, model):
            query = query & scope()
        return query

    def _get_belongs(self, modelname, value):
        return self.model.db[modelname]._model_._belongs_ref_.get(value)

    def belongs_query(self):
        return (self.model.table[self.ref[1]] == self.model.db[self.ref[0]].id)

    @staticmethod
    def many_query(ref, rid):
        if ref.cast and isinstance(rid, _Field):
            rid = rid.cast(ref.cast)
        return ref.model_instance.table[ref.field] == rid

    def _many(self, ref, rid):
        return ref.dbset.where(
            self._patch_query_with_scopes(ref, self.many_query(ref, rid))
        ).query

    def many(self, row=None):
        return self._many(self.ref, self._make_refid(row))

    def via(self, row=None):
        db = self.model.db
        rid = self._make_refid(row)
        sname = self.model.__class__.__name__
        stack = []
        midrel = self.model._hasmany_ref_[self.ref.via]
        stack.append(self.ref)
        while midrel.via is not None:
            stack.insert(0, midrel)
            midrel = self.model._hasmany_ref_[midrel.via]
        query = self._many(midrel, rid)
        step_model = midrel.table_name
        sel_field = db[step_model].ALL
        last_belongs = None
        last_via = None
        for via in stack:
            rname = via.field or via.name[:-1]
            belongs_model = self._get_belongs(step_model, rname)
            if belongs_model:
                #: join table way
                last_belongs = step_model
                last_via = via
                _query = (db[belongs_model].id == db[step_model][rname])
                sel_field = db[belongs_model].ALL
                step_model = belongs_model
            else:
                #: shortcut way
                last_belongs = None
                rname = via.field or via.name
                midrel = db[step_model]._model_._hasmany_ref_[rname]
                _query = self._many(midrel, db[step_model].id)
                step_model = midrel.table_name
                sel_field = db[step_model].ALL
            query = query & _query
        query = via.dbset.where(
            self._patch_query_with_scopes_on(via, query, step_model)).query
        return query, sel_field, sname, rid, last_belongs, last_via


class Callback(object):
    _inst_count_ = 0

    def __init__(self, f, t):
        self.t = []
        if isinstance(f, Callback):
            self.t += f.t
            f = f.f
        self.f = f
        self.t.append(t)
        self._inst_count_ = Callback._inst_count_
        Callback._inst_count_ += 1

    def __call__(self):
        return None


class JoinedIDReference(_IDReference):
    @classmethod
    def _from_record(cls, record, table=None):
        rv = cls(record.id)
        rv._table = table
        rv._record = record
        return rv

    def as_dict(self, datetime_to_str=False, custom_types=None):
        return self._record.as_dict()


class TimingHandler(ExecutionHandler):
    def _timings(self):
        THREAD_LOCAL._emtdal_timings_ = getattr(
            THREAD_LOCAL, '_emtdal_timings_', [])
        return THREAD_LOCAL._emtdal_timings_

    @cachedprop
    def timings(self):
        return self._timings()

    def before_execute(self, command):
        self.t = time.time()

    def after_execute(self, command):
        dt = time.time() - self.t
        self.timings.append((command, dt))


class ConnectionContext:
    __slots__ = ['db', 'conn', 'with_transaction', 'reuse_if_open']

    def __init__(self, db, with_transaction=True, reuse_if_open=True):
        self.db = db
        self.conn = None
        self.with_transaction = with_transaction
        self.reuse_if_open = reuse_if_open

    def __enter__(self):
        self.conn = self.db.connection_open(
            with_transaction=self.with_transaction,
            reuse_if_open=self.reuse_if_open
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.db.connection_close()
        self.conn = None

    async def __aenter__(self):
        self.conn = await self.db.connection_open_loop(
            with_transaction=self.with_transaction,
            reuse_if_open=self.reuse_if_open
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            await self.db.connection_close_loop()
        self.conn = None


def decamelize(name):
    return "_".join(re.findall('[A-Z][^A-Z]*', name)).lower()


def camelize(name):
    return "".join(w.capitalize() for w in name.split("_"))


def make_tablename(classname):
    return decamelize(classname) + "s"


def wrap_scope_on_set(dbset, model_instance, scope):
    @wraps(scope)
    def wrapped(*args, **kwargs):
        return dbset.where(
            scope(model_instance, *args, **kwargs),
            model=model_instance.__class__)
    return wrapped


def wrap_scope_on_model(scope):
    @wraps(scope)
    def wrapped(cls, *args, **kwargs):
        return cls.db.where(
            scope(cls._instance_(), *args, **kwargs), model=cls)
    return wrapped


def wrap_virtual_on_model(model, virtual):
    @wraps(virtual)
    def wrapped(row, *args, **kwargs):
        return virtual(model, row, *args, **kwargs)
    return wrapped



# -*- coding: utf-8 -*-
"""
    emmett.validators.basic
    -----------------------

    Provide basic validators.

    :copyright: 2014 Giovanni Barillari

    Based on the web2py's validators (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import re

from cgi import FieldStorage
from functools import reduce
from os import SEEK_END, SEEK_SET

# TODO: check unicode conversions
from .._shortcuts import to_unicode
from .helpers import translate, is_empty


class Validator:
    message = "Invalid value"

    def __init__(self, message=None):
        self.message = message or self.message

    def formatter(self, value):
        return value

    def __call__(self, value):
        raise NotImplementedError


class ParentValidator(Validator):
    def __init__(self, children, message=None):
        super().__init__(message=message)
        if not isinstance(children, (list, tuple)):
            children = [children]
        self.children = children

    def formatter(self, value):
        return reduce(
            lambda formatted_val, child: child.formatter(formatted_val),
            self.children,
            value
        )

    def __call__(self, value):
        raise NotImplementedError


class _is(Validator):
    message = "Invalid value"
    rule = None

    def __call__(self, value):
        if (
            self.rule is None or (
                self.rule is not None and
                self.rule.match(to_unicode(value) or '')
            )
        ):
            return self.check(value)
        return value, translate(self.message)

    def check(self, value):
        return value, None


class Not(ParentValidator):
    message = "Value not allowed"

    def __call__(self, value):
        val = value
        for child in self.children:
            value, error = child(value)
            if error is None:
                return val, translate(self.message)
        return value, None


class Any(ParentValidator):
    def __call__(self, value):
        for child in self.children:
            value, error = child(value)
            if error is None:
                break
        return value, error


class Allow(ParentValidator):
    def __init__(self, value, children, message=None):
        super().__init__(children, message=message)
        self.value = value

    def __call__(self, value):
        val = value
        comparing = self.value() if callable(self.value) else self.value
        if value is not comparing:
            for child in self.children:
                value, error = child(value)
                if error:
                    return val, error
        return value, None


class isEmpty(Validator):
    message = "No value allowed"

    def __init__(self, empty_regex=None, message=None):
        super().__init__(message=message)
        self.empty_regex = re.compile(empty_regex) if empty_regex is not None else None

    def __call__(self, value):
        value, empty = is_empty(value, empty_regex=self.empty_regex)
        if empty:
            return None, None
        return value, translate(self.message)


class isntEmpty(isEmpty):
    message = "Cannot be empty"

    def __call__(self, value):
        value, empty = is_empty(value, empty_regex=self.empty_regex)
        if empty:
            return value, translate(self.message)
        return value, None


class isEmptyOr(ParentValidator):
    def __init__(self, children, empty_regex=None, message=None):
        super().__init__(children, message=message)
        self.empty_regex = re.compile(empty_regex) if empty_regex is not None else None
        for child in self.children:
            if hasattr(child, 'multiple'):
                self.multiple = child.multiple
                break
        for child in self.children:
            if hasattr(child, 'options'):
                self._options_ = child.options
                self.options = self._get_options_
                break

    def _get_options_(self):
        options = self._options_()
        if (not options or options[0][0] != '') and not self.multiple:
            options.insert(0, ('', ''))
        return options

    def __call__(self, value):
        value, empty = is_empty(value, empty_regex=self.empty_regex)
        if empty:
            return None, None
        error = None
        for child in self.children:
            value, error = child(value)
            if error:
                break
        return value, error


class Equals(Validator):
    message = "No match"

    def __init__(self, expression, message=None):
        super().__init__(message=message)
        self.expression = expression

    def __call__(self, value):
        if value == self.expression:
            return (value, None)
        return value, translate(self.message)


class Matches(Validator):
    message = "Invalid expression"

    def __init__(
        self, expression, strict=False, search=False, extract=False, message=None
    ):
        super().__init__(message=message)
        if strict or not search:
            if not expression.startswith('^'):
                expression = '^(%s)' % expression
        if strict:
            if not expression.endswith('$'):
                expression = '(%s)$' % expression
        self.regex = re.compile(expression)
        self.extract = extract

    def __call__(self, value):
        match = self.regex.search(to_unicode(value) or '')
        if match is not None:
            return self.extract and match.group() or value, None
        return value, translate(self.message)


class hasLength(Validator):
    message = "Enter from {min} to {max} characters"

    def __init__(
        self, maxsize=256, minsize=0, include=(True, False), message=None
    ):
        super().__init__(message=message)
        self.maxsize = maxsize
        self.minsize = minsize
        self.inc = include

    def _between(self, value):
        if self.inc[0]:
            great = self.minsize <= value
        else:
            great = self.minsize < value
        if self.inc[1]:
            less = value <= self.maxsize
        else:
            less = value < self.maxsize
        return great and less

    def __call__(self, value):
        if value is None:
            length = 0
            if self._between(length):
                return value, None
        elif isinstance(value, FieldStorage):
            if value.file:
                value.file.seek(0, SEEK_END)
                length = value.file.tell()
                value.file.seek(0, SEEK_SET)
            elif hasattr(value, 'value'):
                val = value.value
                if val:
                    length = len(val)
                else:
                    length = 0
            if self._between(length):
                return value, None
        elif isinstance(value, bytes):
            try:
                lvalue = len(value.decode('utf8'))
            except Exception:
                lvalue = len(value)
            if self._between(lvalue):
                return value, None
        elif isinstance(value, str):
            if self._between(len(value)):
                return value, None
        elif isinstance(value, (tuple, list)):
            if self._between(len(value)):
                return value, None
        elif self._between(len(str(value))):
            return str(value), None
        return value, translate(self.message).format(min=self.minsize, max=self.maxsize)



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.revisions
    -------------------------------

    Provides revisions logic for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

from collections import deque
from ...datastructures import OrderedSet
from ...utils import cachedprop
from .helpers import to_tuple, tuple_or_value, dedupe_tuple
from .exceptions import (
    RevisionError, RangeNotAncestorError, ResolutionError, MultipleHeads
)


class Revision(object):
    nextrev = frozenset()
    _all_nextrev = frozenset()
    revision = None
    down_revision = None
    #dependencies = None
    #branch_labels = None

    def __init__(self, revision, down_revision,
                 dependencies=None, branch_labels=None):
        self.revision = revision
        self.down_revision = tuple_or_value(down_revision)
        #self.dependencies = tuple_or_value(dependencies)
        #self._resolved_dependencies = ()
        #self._orig_branch_labels = to_tuple(branch_labels, default=())
        #self.branch_labels = set(self._orig_branch_labels)

    def __repr__(self):
        args = [
            repr(self.revision),
            repr(self.down_revision)
        ]
        # if self.dependencies:
        #     args.append("dependencies=%r" % self.dependencies)
        # if self.branch_labels:
        #     args.append("branch_labels=%r" % self.branch_labels)
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(args)
        )

    def add_nextrev(self, revision):
        self._all_nextrev = self._all_nextrev.union([revision.revision])
        if self.revision in revision._versioned_down_revisions:
            self.nextrev = self.nextrev.union([revision.revision])

    @property
    def _all_down_revisions(self):
        return to_tuple(self.down_revision, default=())
        #+ self._resolved_dependencies

    @property
    def _versioned_down_revisions(self):
        return to_tuple(self.down_revision, default=())

    @property
    def is_head(self):
        return not bool(self.nextrev)

    @property
    def _is_real_head(self):
        return not bool(self._all_nextrev)

    @property
    def is_base(self):
        return self.down_revision is None

    # @property
    # def _is_real_base(self):
    #     return self.down_revision is None and self.dependencies is None

    @property
    def is_branch_point(self):
        return len(self.nextrev) > 1

    @property
    def _is_real_branch_point(self):
        return len(self._all_nextrev) > 1

    @property
    def is_merge_point(self):
        return len(self._versioned_down_revisions) > 1


class RevisionsMap(object):
    def __init__(self, app, generator):
        self.app = app
        self._generator = generator

    @cachedprop
    def _revision_map(self):
        rmap = {}

        heads = OrderedSet()
        _real_heads = OrderedSet()
        self.bases = ()
        self._real_bases = ()

        #has_branch_labels = set()
        #has_depends_on = set()
        for revision in self._generator():

            if revision.revision in rmap:
                self.app.log.warn(
                    "Revision %s is present more than once" % revision.revision
                )
            rmap[revision.revision] = revision
            # if revision.branch_labels:
            #     has_branch_labels.add(revision)
            # if revision.dependencies:
            #     has_depends_on.add(revision)
            heads.add(revision.revision)
            _real_heads.add(revision.revision)
            if revision.is_base:
                self.bases += (revision.revision, )
                self._real_bases += (revision.revision, )
            #if revision._is_real_base:
            #    self._real_bases += (revision.revision, )

        # for revision in has_branch_labels:
        #     self._map_branch_labels(revision, rmap)

        # for revision in has_depends_on:
        #     self._add_depends_on(revision, rmap)

        for rev in rmap.values():
            for downrev in rev._all_down_revisions:
                if downrev not in rmap:
                    self.app.log.warn(
                        "Revision %s referenced from %s is not present" %
                        (downrev, rev))
                down_revision = rmap[downrev]
                down_revision.add_nextrev(rev)
                if downrev in rev._versioned_down_revisions:
                    heads.discard(downrev)
                _real_heads.discard(downrev)

        rmap[None] = rmap[()] = None
        self.heads = tuple(heads)
        self._real_heads = tuple(_real_heads)

        # for revision in has_branch_labels:
        #     self._add_branches(revision, rmap, map_branch_labels=False)
        return rmap

    @cachedprop
    def heads(self):
        self._revision_map
        return self.heads

    @cachedprop
    def bases(self):
        self._revision_map
        return self.bases

    @cachedprop
    def _real_heads(self):
        self._revision_map
        return self._real_heads

    @cachedprop
    def _real_bases(self):
        self._revision_map
        return self._real_bases

    def get_current_head(self):
        current_heads = self.heads
        if len(current_heads) > 1:
            raise MultipleHeads(current_heads, "head")
        if current_heads:
            return current_heads[0]
        else:
            return None

    def _resolve_revision_number(self, rid):
        self._revision_map
        if rid == 'heads':
            return self._real_heads
        elif rid == 'head':
            current_head = self.get_current_head()
            if current_head:
                return (current_head, )
            return ()
        elif rid == 'base' or rid is None:
            return ()
        else:
            return to_tuple(rid, default=None)

    def _revision_for_ident(self, resolved_id):
        try:
            revision = self._revision_map[resolved_id]
        except KeyError:
            # do a partial lookup
            revs = [x for x in self._revision_map
                    if x and x.startswith(resolved_id)]
            if not revs:
                raise ResolutionError(
                    "No such revision or branch '%s'" % resolved_id,
                    resolved_id)
            elif len(revs) > 1:
                raise ResolutionError(
                    "Multiple revisions start "
                    "with '%s': %s..." % (
                        resolved_id,
                        ", ".join("'%s'" % r for r in revs[0:3])
                    ), resolved_id)
            else:
                revision = self._revision_map[revs[0]]
        return revision

    def get_revision(self, rid):
        resolved_id = self._resolve_revision_number(rid)
        if len(resolved_id) > 1:
            raise MultipleHeads(resolved_id, rid)
        elif resolved_id:
            resolved_id = resolved_id[0]
        return self._revision_for_ident(resolved_id)

    def get_revisions(self, rid):
        #: return revision instances with the given rev id or identifiers
        if isinstance(rid, (list, tuple, set, frozenset)):
            return sum([self.get_revisions(id_elem) for id_elem in rid], ())
        else:
            resolved_id = self._resolve_revision_number(rid)
            return tuple(
                self._revision_for_ident(rev_id) for rev_id in resolved_id)

    def add_revision(self, revision, _replace=False):
        map_ = self._revision_map
        if not _replace and revision.revision in map_:
            self.app.log.warn(
                "Revision %s is present more than once" % revision.revision)
        elif _replace and revision.revision not in map_:
            raise Exception("revision %s not in map" % revision.revision)

        map_[revision.revision] = revision

        if revision.is_base:
            self.bases += (revision.revision, )
            self._real_bases += (revision.revision, )
        # if revision._is_real_base:
        #     self._real_bases += (revision.revision, )
        for downrev in revision._all_down_revisions:
            if downrev not in map_:
                self.app.log.warn(
                    "Revision %s referenced from %s is not present"
                    % (downrev, revision)
                )
            map_[downrev].add_nextrev(revision)
        if revision._is_real_head:
            self._real_heads = tuple(
                head for head in self._real_heads
                if head not in
                set(revision._all_down_revisions).union([revision.revision])
            ) + (revision.revision,)
        if revision.is_head:
            self.heads = tuple(
                head for head in self.heads
                if head not in
                set(revision._versioned_down_revisions).union(
                    [revision.revision])
            ) + (revision.revision,)

    def iterate_revisions(self, upper, lower, implicit_base=False,
                          inclusive=False):
        #: iterate through script revisions, starting at the given upper
        #  revision identifier and ending at the lower.
        return self._iterate_revisions(
            upper, lower, inclusive=inclusive, implicit_base=implicit_base)

    def _get_ancestor_nodes(self, targets, map_=None, check=False):
        fn = lambda rev: rev._versioned_down_revisions

        return self._iterate_related_revisions(
            fn, targets, map_=map_, check=check
        )

    def _get_descendant_nodes(self, targets, map_=None, check=False):
        fn = lambda rev: rev.nextrev

        return self._iterate_related_revisions(
            fn, targets, map_=map_, check=check
        )

    def _iterate_related_revisions(self, fn, targets, map_, check=False):
        if map_ is None:
            map_ = self._revision_map

        seen = set()
        todo = deque()
        for target in targets:

            todo.append(target)
            if check:
                per_target = set()

            while todo:
                rev = todo.pop()
                if check:
                    per_target.add(rev)

                if rev in seen:
                    continue
                seen.add(rev)
                todo.extend(
                    map_[rev_id] for rev_id in fn(rev))
                yield rev
            if check and per_target.intersection(targets).difference([target]):
                raise RevisionError(
                    "Requested revision %s overlaps with "
                    "other requested revisions" % target.revision)

    def _iterate_revisions(self, upper, lower, inclusive=True,
                           implicit_base=False):
        #: iterate revisions from upper to lower.
        requested_lowers = self.get_revisions(lower)
        uppers = dedupe_tuple(self.get_revisions(upper))

        if not uppers and not requested_lowers:
            return

        upper_ancestors = set(self._get_ancestor_nodes(uppers, check=True))

        if implicit_base and requested_lowers:
            lower_ancestors = set(
                self._get_ancestor_nodes(requested_lowers)
            )
            lower_descendants = set(
                self._get_descendant_nodes(requested_lowers)
            )
            base_lowers = set()
            candidate_lowers = upper_ancestors.\
                difference(lower_ancestors).\
                difference(lower_descendants)
            for rev in candidate_lowers:
                for downrev in rev._all_down_revisions:
                    if self._revision_map[downrev] in candidate_lowers:
                        break
                else:
                    base_lowers.add(rev)
            lowers = base_lowers.union(requested_lowers)
        elif implicit_base:
            base_lowers = set(self.get_revisions(self._real_bases))
            lowers = base_lowers.union(requested_lowers)
        elif not requested_lowers:
            lowers = set(self.get_revisions(self._real_bases))
        else:
            lowers = requested_lowers

        # represents all nodes we will produce
        total_space = set(
            rev.revision for rev in upper_ancestors).intersection(
            rev.revision for rev
            in self._get_descendant_nodes(lowers, check=True)
        )

        if not total_space:
            raise RangeNotAncestorError(lower, upper)

        # organize branch points to be consumed separately from
        # member nodes
        branch_todo = set(
            rev for rev in
            (self._revision_map[rev] for rev in total_space)
            if rev._is_real_branch_point and
            len(total_space.intersection(rev._all_nextrev)) > 1
        )

        # it's not possible for any "uppers" to be in branch_todo,
        # because the ._all_nextrev of those nodes is not in total_space
        #assert not branch_todo.intersection(uppers)

        todo = deque(
            r for r in uppers if r.revision in total_space)

        # iterate for total_space being emptied out
        total_space_modified = True
        while total_space:

            if not total_space_modified:
                raise RevisionError(
                    "Dependency resolution failed; iteration can't proceed")
            total_space_modified = False
            # when everything non-branch pending is consumed,
            # add to the todo any branch nodes that have no
            # descendants left in the queue
            if not todo:
                todo.extendleft(
                    sorted(
                        (
                            rev for rev in branch_todo
                            if not rev._all_nextrev.intersection(total_space)
                        ),
                        # favor "revisioned" branch points before
                        # dependent ones
                        key=lambda rev: 0 if rev.is_branch_point else 1
                    )
                )
                branch_todo.difference_update(todo)
            # iterate nodes that are in the immediate todo
            while todo:
                rev = todo.popleft()
                total_space.remove(rev.revision)
                total_space_modified = True

                # do depth first for elements within branches,
                # don't consume any actual branch nodes
                todo.extendleft([
                    self._revision_map[downrev]
                    for downrev in reversed(rev._all_down_revisions)
                    if self._revision_map[downrev] not in branch_todo and
                    downrev in total_space])

                if not inclusive and rev in requested_lowers:
                    continue
                yield rev

        assert not branch_todo



# -*- coding: utf-8 -*-
"""
    tests.cache
    -----------

    Test Emmett cache module
"""

import pytest

from collections import defaultdict

from emmett import App
from emmett.cache import CacheHandler, RamCache, DiskCache, Cache


def test_basecache():
    base_cache = CacheHandler()
    assert base_cache._default_expire == 300

    assert base_cache('key', lambda: 'value') is 'value'
    assert base_cache.get('key') is None
    assert base_cache.set('key', 'value', 300) is None
    assert base_cache.clear() is None


async def _await_2():
    return 2


async def _await_3():
    return 3


@pytest.mark.asyncio
async def test_ramcache():
    ram_cache = RamCache()
    assert ram_cache._prefix == ''
    assert ram_cache._threshold == 500

    assert ram_cache('test', lambda: 2) == 2
    assert ram_cache('test', lambda: 3, 300) == 2

    assert await ram_cache('test_loop', _await_2) == 2
    assert await ram_cache('test_loop', _await_3, 300) == 2

    ram_cache.set('test', 3)
    assert ram_cache.get('test') == 3

    ram_cache.set('test', 4, 300)
    assert ram_cache.get('test') == 4

    ram_cache.clear()
    assert ram_cache.get('test') is None


@pytest.mark.asyncio
async def test_diskcache():
    App(__name__)

    disk_cache = DiskCache()
    assert disk_cache._threshold == 500

    assert disk_cache('test', lambda: 2) == 2
    assert disk_cache('test', lambda: 3, 300) == 2

    assert await disk_cache('test_loop', _await_2) == 2
    assert await disk_cache('test_loop', _await_3, 300) == 2

    disk_cache.set('test', 3)
    assert disk_cache.get('test') == 3

    disk_cache.set('test', 4, 300)
    assert disk_cache.get('test') == 4

    disk_cache.clear()
    assert disk_cache.get('test') is None


@pytest.mark.asyncio
async def test_cache():
    default_cache = Cache()
    assert isinstance(default_cache._default_handler, RamCache)

    assert default_cache('test', lambda: 2) == 2
    assert default_cache('test', lambda: 3) == 2

    assert await default_cache('test_loop', _await_2) == 2
    assert await default_cache('test_loop', _await_3, 300) == 2

    default_cache.set('test', 3)
    assert default_cache('test', lambda: 2) == 3

    default_cache.set('test', 4, 300)
    assert default_cache('test', lambda: 2, 300) == 4

    default_cache.clear()

    disk_cache = DiskCache()
    ram_cache = RamCache()
    cache = Cache(default='disc', ram=ram_cache, disc=disk_cache)
    assert isinstance(cache._default_handler, DiskCache)
    assert cache.disc == disk_cache
    assert cache.ram == ram_cache


def test_cache_decorator_sync():
    cache = Cache(ram=RamCache(prefix='test:'))
    calls = defaultdict(lambda: 0)

    @cache('foo')
    def foo(*args, **kwargs):
        calls['foo'] += 1
        return 'foo'

    #: no arguments
    for _ in range(0, 2):
        foo()
    assert len(cache._default_handler.data.keys()) == 1
    assert calls['foo'] == 1

    #: args change the cache key
    for _ in range(0, 2):
        foo(1)
    assert len(cache._default_handler.data.keys()) == 2
    assert calls['foo'] == 2

    for _ in range(0, 2):
        foo(1, 2)
    assert len(cache._default_handler.data.keys()) == 3
    assert calls['foo'] == 3

    #: kwargs change the cache key
    for _ in range(0, 2):
        foo(1, a='foo', b='bar')
    assert len(cache._default_handler.data.keys()) == 4
    assert calls['foo'] == 4

    #: kwargs order won't change the cache key
    for _ in range(0, 2):
        foo(1, b='bar', a='foo')
    assert len(cache._default_handler.data.keys()) == 4
    assert calls['foo'] == 4


@pytest.mark.asyncio
async def test_cache_decorator_loop():
    cache = Cache(ram=RamCache(prefix='bar:'))
    calls = defaultdict(lambda: 0)

    @cache('bar')
    async def bar(*args, **kwargs):
        calls['bar'] += 1
        return 'bar'

    #: no arguments
    for _ in range(0, 2):
        await bar()
    assert len(cache._default_handler.data.keys()) == 1
    assert calls['bar'] == 1

    #: args change the cache key
    for _ in range(0, 2):
        await bar(1)
    assert len(cache._default_handler.data.keys()) == 2
    assert calls['bar'] == 2

    for _ in range(0, 2):
        await bar(1, 2)
    assert len(cache._default_handler.data.keys()) == 3
    assert calls['bar'] == 3

    #: kwargs change the cache key
    for _ in range(0, 2):
        await bar(1, a='foo', b='bar')
    assert len(cache._default_handler.data.keys()) == 4
    assert calls['bar'] == 4

    #: kwargs order won't change the cache key
    for _ in range(0, 2):
        await bar(1, b='bar', a='foo')
    assert len(cache._default_handler.data.keys()) == 4
    assert calls['bar'] == 4



# -*- coding: utf-8 -*-
"""
    tests.templates
    ---------------

    Test Emmett templating module
"""

import pytest

from helpers import current_ctx
from emmett import App


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    app.config.templates_escape = 'all'
    app.config.templates_adjust_indent = True
    app.config.templates_auto_reload = True
    return app


def test_helpers(app):
    templater = app.templater
    r = templater._render(source="{{ include_helpers }}")
    assert r == '<script type="text/javascript" ' + \
        'src="/__emmett__/jquery.min.js"></script>' + \
        '<script type="text/javascript" ' + \
        'src="/__emmett__/helpers.js"></script>'


def test_meta(app):
    with current_ctx('/', app) as ctx:
        ctx.response.meta.foo = "bar"
        ctx.response.meta_prop.foo = "bar"
        templater = app.templater
        r = templater._render(
            source="{{ include_meta }}",
            context={'current': ctx})
        assert r == '<meta name="foo" content="bar" />' + \
            '<meta property="foo" content="bar" />'


def test_static(app):
    templater = app.templater
    s = "{{include_static 'foo.js'}}\n{{include_static 'bar.css'}}"
    r = templater._render(source=s)
    assert r == '<script type="text/javascript" src="/static/foo.js">' + \
        '</script>\n<link rel="stylesheet" href="/static/bar.css" ' + \
        'type="text/css" />'


rendered_value = """
<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>

        {helpers}
        <link rel="stylesheet" href="/static/style.css" type="text/css" />
    </head>
    <body>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">nuvolosit&#224; variabile</a>
            </div>

            <ul class="posts">
                <li>
                    <h2>foo</h2>
                    <hr />
                </li>
                <li>
                    <h2>bar</h2>
                    <hr />
                </li>
            </ul>

        </div>
    </body>
</html>""".format(
    helpers="".join([
        f'<script type="text/javascript" src="/__emmett__/{name}"></script>'
        for name in ["jquery.min.js", "helpers.js"]
    ])
)


def test_render(app):
    with current_ctx('/', app) as ctx:
        ctx.language = 'it'
        r = app.templater.render(
            'test.html', {
                'current': ctx, 'posts': [{'title': 'foo'}, {'title': 'bar'}]
            }
        )
        assert "\n".join([l.rstrip() for l in r.splitlines()]) == \
            rendered_value[1:]



# -*- coding: utf-8 -*-
"""
    emmett.routing.dispatchers
    --------------------------

    Provides pipeline dispatchers for routes.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from ..ctx import current


class Dispatcher:
    __slots__ = ['f', 'flow_open', 'flow_close']

    def __init__(self, route):
        self.f = route.f
        self.flow_open = route.pipeline_flow_open
        self.flow_close = route.pipeline_flow_close

    async def _parallel_flow(self, flow):
        tasks = [asyncio.create_task(method()) for method in flow]
        await asyncio.gather(*tasks, return_exceptions=True)
        for task in tasks:
            if task.exception():
                raise task.exception()

    def dispatch(self, reqargs):
        return self.f(**reqargs)


class RequestDispatcher(Dispatcher):
    __slots__ = ['response_builder']

    def __init__(self, route, rule, response_builder):
        super().__init__(route)
        self.response_builder = response_builder

    async def dispatch(self, reqargs, response):
        return self.response_builder(await self.f(**reqargs), response)


class RequestOpenDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        return self.response_builder(await self.f(**reqargs), response)


class RequestCloseDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        try:
            rv = self.response_builder(await self.f(**reqargs), response)
        finally:
            await self._parallel_flow(self.flow_close)
        return rv


class RequestFlowDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        try:
            rv = self.response_builder(await self.f(**reqargs), response)
        finally:
            await self._parallel_flow(self.flow_close)
        return rv


class WSOpenDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        await self._parallel_flow(self.flow_open)
        await self.f(**reqargs)


class WSCloseDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        try:
            await self.f(**reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))


class WSFlowDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        await self._parallel_flow(self.flow_open)
        try:
            await self.f(**reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))


class CacheDispatcher(RequestDispatcher):
    __slots__ = ['route', 'cache_rule']

    def __init__(self, route, rule, response_builder):
        super().__init__(route, rule, response_builder)
        self.route = route
        self.cache_rule = rule.cache_rule

    async def get_data(self, reqargs, response):
        key = self.cache_rule._build_ctx_key(
            self.route, **self.cache_rule._build_ctx(
                reqargs, self.route, current
            )
        )
        data = self.cache_rule.cache.get(key)
        if data is not None:
            response.headers.update(data['headers'])
            return data['content']
        content = await self.f(**reqargs)
        if response.status == 200:
            self.cache_rule.cache.set(
                key,
                {'content': content, 'headers': response.headers},
                self.cache_rule.duration
            )
        return content

    async def dispatch(self, reqargs, response):
        content = await self.get_data(reqargs, response)
        return self.response_builder(content, response)


class CacheOpenDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        return await super().dispatch(reqargs, response)


class CacheCloseDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        try:
            content = await self.get_data(reqargs, response)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return self.response_builder(content, response)


class CacheFlowDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        try:
            content = await self.get_data(reqargs, response)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return self.response_builder(content, response)
